from typing import Dict
from typing import Iterator
from typing import List
from typing import Optional
from typing import Tuple

from xsdata.codegen.models import Attr
from xsdata.codegen.models import AttrType
from xsdata.codegen.models import Class
from xsdata.codegen.models import Extension
from xsdata.codegen.models import Restrictions
from xsdata.models.enums import DataType
from xsdata.models.enums import Tag
from xsdata.models.mixins import ElementBase
from xsdata.models.xsd import Attribute
from xsdata.models.xsd import AttributeGroup
from xsdata.models.xsd import ComplexType
from xsdata.models.xsd import Element
from xsdata.models.xsd import Group
from xsdata.models.xsd import Schema
from xsdata.models.xsd import SimpleType
from xsdata.utils import collections
from xsdata.utils import text
from xsdata.utils.namespaces import build_qname
from xsdata.utils.namespaces import is_default
from xsdata.utils.namespaces import prefix_exists


class SchemaMapper:
    """Map a schema instance to classes, extensions and attributes."""

    @classmethod
    def map(cls, schema: Schema) -> List[Class]:
        """Map schema children elements to classes."""

        assert schema.location is not None

        location = schema.location
        target_namespace = schema.target_namespace

        return [
            cls.build_class(element, container, location, target_namespace)
            for container, element in cls.root_elements(schema)
        ]

    @classmethod
    def root_elements(cls, schema: Schema):
        """Return all valid schema elements that can be converted to
        classes."""

        for override in schema.overrides:
            for child in override.children(condition=cls.is_class):
                yield Tag.OVERRIDE, child

        for redefine in schema.redefines:
            for child in redefine.children(condition=cls.is_class):
                yield Tag.REDEFINE, child

        for child in schema.children(condition=cls.is_class):
            yield Tag.SCHEMA, child

    @classmethod
    def build_class(
        cls,
        obj: ElementBase,
        container: str,
        location: str,
        target_namespace: Optional[str],
    ) -> Class:
        """Build and return a class instance."""
        instance = Class(
            qname=build_qname(target_namespace, obj.real_name),
            abstract=obj.is_abstract,
            namespace=cls.element_namespace(obj, target_namespace),
            mixed=obj.is_mixed,
            nillable=obj.is_nillable,
            tag=obj.class_name,
            container=container,
            help=obj.display_help,
            ns_map=obj.ns_map,
            location=location,
            default=obj.default_value,
            fixed=obj.is_fixed,
            substitutions=cls.build_substitutions(obj, target_namespace),
        )

        cls.build_class_extensions(obj, instance)
        cls.build_class_attributes(obj, instance)
        return instance

    @classmethod
    def build_substitutions(
        cls, obj: ElementBase, target_namespace: Optional[str]
    ) -> List[str]:
        return [
            build_qname(obj.ns_map.get(prefix, target_namespace), suffix)
            for prefix, suffix in map(text.split, obj.substitutions)
        ]

    @classmethod
    def build_class_attributes(cls, obj: ElementBase, target: Class):
        """Build the target class attributes from the given ElementBase
        children."""

        base_restrictions = Restrictions.from_element(obj)
        for child, restrictions in cls.element_children(obj, base_restrictions):
            cls.build_class_attribute(target, child, restrictions)

        target.attrs.sort(key=lambda x: x.index)

    @classmethod
    def build_class_extensions(cls, obj: ElementBase, target: Class):
        """Build the item class extensions from the given ElementBase
        children."""

        restrictions = obj.get_restrictions()
        extensions = [
            cls.build_class_extension(target, base, restrictions) for base in obj.bases
        ]
        extensions.extend(cls.children_extensions(obj, target))
        target.extensions = collections.unique_sequence(extensions)

    @classmethod
    def build_data_type(
        cls, target: Class, name: str, forward: bool = False
    ) -> AttrType:
        """Create an attribute type for the target class."""
        prefix, suffix = text.split(name)
        namespace = target.ns_map.get(prefix, target.target_namespace)
        qname = build_qname(namespace, suffix)
        datatype = DataType.from_qname(qname)

        return AttrType(
            qname=qname,
            native=datatype is not None,
            forward=forward,
        )

    @classmethod
    def element_children(
        cls, obj: ElementBase, parent_restrictions: Restrictions
    ) -> Iterator[Tuple[ElementBase, Restrictions]]:
        """Recursively find and return all child elements that are qualified to
        be class attributes, with all their restrictions."""

        for child in obj.children():
            if child.is_property:
                yield child, parent_restrictions
            else:
                restrictions = parent_restrictions.clone()
                restrictions.merge(Restrictions.from_element(child))
                yield from cls.element_children(child, restrictions)

    @classmethod
    def element_namespace(
        cls, obj: ElementBase, target_namespace: Optional[str]
    ) -> Optional[str]:
        """
        Return the target namespace for the given schema element.

        In order:
            - elements/attributes with specific target namespace
            - prefixed elements returns the namespace from schema ns_map
            - qualified elements returns the schema target namespace
            - unqualified elements return an empty string
            - unqualified attributes return None
        """

        raw_namespace = obj.raw_namespace
        if raw_namespace:
            return raw_namespace

        prefix = obj.prefix
        if prefix:
            return obj.ns_map.get(prefix)

        if obj.is_qualified and (
            not obj.is_ref
            or not target_namespace
            or not prefix_exists(target_namespace, obj.ns_map)
            or is_default(target_namespace, obj.ns_map)
        ):
            return target_namespace

        return "" if isinstance(obj, Element) else None

    @classmethod
    def children_extensions(
        cls, obj: ElementBase, target: Class
    ) -> Iterator[Extension]:
        """
        Recursively find and return all target's Extension classes.

        If the initial given obj has a type attribute include it in
        result.
        """
        for child in obj.children():
            if child.is_property:
                continue

            for ext in child.bases:
                yield cls.build_class_extension(target, ext, child.get_restrictions())

            yield from cls.children_extensions(child, target)

    @classmethod
    def build_class_extension(
        cls, target: Class, name: str, restrictions: Dict
    ) -> Extension:
        """Create an extension for the target class."""
        return Extension(
            type=cls.build_data_type(target, name),
            restrictions=Restrictions(**restrictions),
        )

    @classmethod
    def build_class_attribute(
        cls,
        target: Class,
        obj: ElementBase,
        parent_restrictions: Restrictions,
    ):
        """Generate and append an attribute field to the target class."""

        target.ns_map.update(obj.ns_map)
        types = cls.build_class_attribute_types(target, obj)
        restrictions = Restrictions.from_element(obj)

        if obj.class_name in (Tag.ELEMENT, Tag.ANY, Tag.GROUP):
            restrictions.merge(parent_restrictions)

        if restrictions.is_prohibited:
            return

        name = obj.real_name
        target.attrs.append(
            Attr(
                index=obj.index,
                name=name,
                default=obj.default_value,
                fixed=obj.is_fixed,
                types=types,
                tag=obj.class_name,
                help=obj.display_help,
                namespace=cls.element_namespace(obj, target.target_namespace),
                restrictions=restrictions,
            )
        )

    @classmethod
    def build_class_attribute_types(
        cls, target: Class, obj: ElementBase
    ) -> List[AttrType]:
        """Convert real type and anonymous inner types to an attribute type
        list."""

        types = [cls.build_data_type(target, tp) for tp in obj.attr_types]

        location = target.location
        namespace = target.target_namespace
        for inner in cls.build_inner_classes(obj, location, namespace):
            target.inner.append(inner)
            types.append(AttrType(qname=inner.qname, forward=True))

        if len(types) == 0:
            types.append(cls.build_data_type(target, name=obj.default_type))

        return collections.unique_sequence(types)

    @classmethod
    def build_inner_classes(
        cls, obj: ElementBase, location: str, namespace: Optional[str]
    ) -> Iterator[Class]:
        """Find and convert anonymous types to a class instances."""
        if isinstance(obj, SimpleType) and obj.is_enumeration:
            yield cls.build_class(obj, obj.class_name, location, namespace)
        else:
            for child in obj.children():
                if isinstance(child, ComplexType) or (
                    isinstance(child, SimpleType) and child.is_enumeration
                ):
                    child.name = obj.real_name
                    yield cls.build_class(child, obj.class_name, location, namespace)
                else:
                    yield from cls.build_inner_classes(child, location, namespace)

    @classmethod
    def is_class(cls, item: ElementBase) -> bool:
        return isinstance(
            item, (SimpleType, ComplexType, Group, AttributeGroup, Element, Attribute)
        )



from unittest import TestCase

from xsdata.models.xsd import Alternative


class AlternativeTests(TestCase):
    def test_property_real_name(self):
        obj = Alternative()
        self.assertEqual("value", obj.real_name)

        obj.id = "foo"
        self.assertEqual("foo", obj.real_name)

        obj.test = "@type='text'"
        self.assertEqual("type_text", obj.real_name)

    def test_property_bases(self):
        obj = Alternative()
        self.assertEqual([], list(obj.bases))

        obj.type = "foo"
        self.assertEqual(["foo"], list(obj.bases))

    def test_get_restrictions(self):
        obj = Alternative()
        expected = {"min_occurs": 0, "choice": str(id(obj))}
        self.assertEqual(expected, obj.get_restrictions())



from dataclasses import dataclass
from dataclasses import field
from typing import Dict
from typing import List
from typing import Optional
from typing import Type

from xsdata.formats.dataclass.models.elements import XmlType
from xsdata.formats.dataclass.models.elements import XmlVar
from xsdata.formats.dataclass.parsers.bases import NodeParser
from xsdata.formats.dataclass.parsers.bases import Parsed
from xsdata.formats.dataclass.parsers.handlers import default_handler
from xsdata.formats.dataclass.parsers.mixins import XmlHandler
from xsdata.formats.dataclass.parsers.mixins import XmlNode
from xsdata.formats.dataclass.parsers.nodes.wildcard import WildcardNode


@dataclass
class TreeParser(NodeParser):
    """
    Bind xml nodes to a tree of AnyElement objects.

    :param handler: Override default XmlHandler
    """

    handler: Type[XmlHandler] = field(default=default_handler())

    def start(
        self,
        clazz: Optional[Type],
        queue: List[XmlNode],
        objects: List[Parsed],
        qname: str,
        attrs: Dict,
        ns_map: Dict,
    ):
        try:
            item = queue[-1]
            child = item.child(qname, attrs, ns_map, len(objects))
        except IndexError:
            var = XmlVar(
                name=qname,
                qname=qname,
                xml_type=XmlType.WILDCARD,
                index=0,
                types=(object,),
                clazz=None,
                init=True,
                mixed=False,
                factory=None,
                tokens_factory=None,
                format=None,
                derived=False,
                any_type=False,
                required=False,
                nillable=False,
                sequential=False,
                default=None,
                namespaces=(),
                elements={},
                wildcards=(),
            )

            child = WildcardNode(
                var=var,
                attrs=attrs,
                ns_map=ns_map,
                position=0,
                factory=self.context.class_type.any_element,
            )
        queue.append(child)



import sys
from dataclasses import dataclass
from dataclasses import field
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Type
from typing import Union
from urllib.parse import urljoin

from xsdata.formats.bindings import T
from xsdata.formats.dataclass.parsers.bases import Parsed
from xsdata.formats.dataclass.parsers.mixins import XmlNode
from xsdata.formats.dataclass.parsers.xml import UserXmlParser
from xsdata.models import xsd
from xsdata.models.enums import FormType
from xsdata.models.enums import Mode
from xsdata.models.enums import Namespace
from xsdata.models.mixins import ElementBase

OPEN_CONTENT_ELEMENT = Union[xsd.ComplexType, xsd.Restriction, xsd.Extension]


@dataclass
class SchemaParser(UserXmlParser):
    """
    A simple parser to convert an xsd schema to an easy to handle data
    structure based on dataclasses.

    The parser is a dummy as possible but it will try to normalize
    certain things like apply parent properties to children.

    :param location:
    :param element_form:
    :param attribute_form:
    :param target_namespace:
    :param default_attributes:
    :param default_open_content:
    """

    index: int = field(default_factory=int)
    indices: List[int] = field(default_factory=list)
    location: Optional[str] = field(default=None)
    element_form: Optional[FormType] = field(init=False, default=None)
    attribute_form: Optional[FormType] = field(init=False, default=None)
    target_namespace: Optional[str] = field(default=None)
    default_attributes: Optional[str] = field(default=None)
    default_open_content: Optional[xsd.DefaultOpenContent] = field(default=None)

    def start(
        self,
        clazz: Optional[Type[T]],
        queue: List[XmlNode],
        objects: List[Parsed],
        qname: str,
        attrs: Dict,
        ns_map: Dict,
    ):
        self.index += 1
        self.indices.append(self.index)
        super().start(clazz, queue, objects, qname, attrs, ns_map)

    def end(
        self,
        queue: List[XmlNode],
        objects: List[Parsed],
        qname: str,
        text: Optional[str],
        tail: Optional[str],
    ) -> Any:
        """Override parent method to set element index and namespaces map."""
        item = queue[-1]
        super().end(queue, objects, qname, text, tail)

        obj = objects[-1][1]
        self.set_index(obj, self.indices.pop())
        self.set_namespace_map(obj, getattr(item, "ns_map", None))

        return obj

    def start_schema(self, attrs: Dict):
        """Collect the schema's default form for attributes and elements for
        later usage."""

        self.element_form = attrs.get("elementFormDefault", None)
        self.attribute_form = attrs.get("attributeFormDefault", None)
        self.default_attributes = attrs.get("defaultAttributes", None)

    def end_schema(self, obj: T):
        """Normalize various properties for the schema and it's children."""
        if isinstance(obj, xsd.Schema):
            self.set_schema_forms(obj)
            self.set_schema_namespaces(obj)
            self.add_default_imports(obj)
            self.resolve_schemas_locations(obj)
            self.reset_element_occurs(obj)

    def end_attribute(self, obj: T):
        """Assign the schema's default form for attributes if the given
        attribute form is None."""
        if isinstance(obj, xsd.Attribute) and obj.form is None and self.attribute_form:
            obj.form = FormType(self.attribute_form)

    def end_complex_type(self, obj: T):
        """
        Post parsing processor to apply default open content and attributes if
        applicable.

        Default open content doesn't apply if the current complex type
        has one of complex content, simple content or has its own open
        content.
        """
        if not isinstance(obj, xsd.ComplexType):
            return

        if obj.default_attributes_apply and self.default_attributes:
            attribute_group = xsd.AttributeGroup(ref=self.default_attributes)
            obj.attribute_groups.insert(0, attribute_group)

        if (
            obj.simple_content
            or obj.complex_content
            or obj.open_content
            or not self.default_open_content
        ):
            return

        if self.default_open_content.applies_to_empty or self.has_elements(obj):
            obj.open_content = self.default_open_content

    def end_default_open_content(self, obj: T):
        """Set the instance default open content to be used later as a property
        for all extensions and restrictions."""
        if isinstance(obj, xsd.DefaultOpenContent):
            if obj.any and obj.mode == Mode.SUFFIX:
                obj.any.index = sys.maxsize

            self.default_open_content = obj

    def end_element(self, obj: T):
        """Assign the schema's default form for elements if the given element
        form is None."""
        if isinstance(obj, xsd.Element) and obj.form is None and self.element_form:
            obj.form = FormType(self.element_form)

    def end_extension(self, obj: T):
        """Set the open content if any to the given extension."""
        if isinstance(obj, xsd.Extension) and not obj.open_content:
            obj.open_content = self.default_open_content

    @classmethod
    def end_open_content(cls, obj: T):
        """Adjust the index to trick later processors into putting attributes
        derived from this open content last in classes."""
        if isinstance(obj, xsd.OpenContent) and obj.any and obj.mode == Mode.SUFFIX:
            obj.any.index = sys.maxsize

    def end_restriction(self, obj: T):
        """Set the open content if any to the given restriction."""
        if isinstance(obj, xsd.Restriction) and not obj.open_content:
            obj.open_content = self.default_open_content

    def set_schema_forms(self, obj: xsd.Schema):
        """
        Set the default form type for elements and attributes.

        Global elements and attributes are by default qualified.
        """
        if self.element_form:
            obj.element_form_default = FormType(self.element_form)
        if self.attribute_form:
            obj.attribute_form_default = FormType(self.attribute_form)

        for child_element in obj.elements:
            child_element.form = FormType.QUALIFIED

        for child_attribute in obj.attributes:
            child_attribute.form = FormType.QUALIFIED

    def set_schema_namespaces(self, obj: xsd.Schema):
        """Set the given schema's target namespace and add the default
        namespaces if the are missing xsi, xlink, xml, xs."""
        obj.target_namespace = obj.target_namespace or self.target_namespace

    def resolve_schemas_locations(self, obj: xsd.Schema):
        """Resolve the locations of the schema overrides, redefines, includes
        and imports relatively to the schema location."""
        if not self.location:
            return

        obj.location = self.location
        for over in obj.overrides:
            over.location = self.resolve_path(over.schema_location)

        for red in obj.redefines:
            red.location = self.resolve_path(red.schema_location)

        for inc in obj.includes:
            inc.location = self.resolve_path(inc.schema_location)

        for imp in obj.imports:
            imp.location = self.resolve_local_path(imp.schema_location, imp.namespace)

    def resolve_path(self, location: Optional[str]) -> Optional[str]:
        """Resolve the given location string relatively the schema location
        path."""

        return urljoin(self.location, location) if self.location and location else None

    def resolve_local_path(
        self, location: Optional[str], namespace: Optional[str]
    ) -> Optional[str]:
        """Resolve the given namespace to one of the local standard schemas or
        fallback to the external file path."""

        common_ns = Namespace.get_enum(namespace)
        local_path = common_ns.location if common_ns else None

        if local_path and (not location or location.find("w3.org/") > 0):
            return local_path

        return self.resolve_path(location)

    @classmethod
    def has_elements(cls, obj: ElementBase) -> bool:
        accepted_types = (xsd.Element, xsd.Any, xsd.Group)
        return any(
            isinstance(child, accepted_types) or cls.has_elements(child)
            for child in obj.children()
        )

    @classmethod
    def set_namespace_map(cls, obj: Any, ns_map: Optional[Dict]):
        """Add common namespaces like xml, xsi, xlink if they are missing."""
        if hasattr(obj, "ns_map"):

            if ns_map:
                obj.ns_map.update(
                    {prefix: uri for prefix, uri in ns_map.items() if uri}
                )

            ns_list = obj.ns_map.values()
            obj.ns_map.update(
                {
                    ns.prefix: ns.uri
                    for ns in Namespace.common()
                    if ns.uri not in ns_list
                }
            )

    @classmethod
    def set_index(cls, obj: Any, index: int):
        if hasattr(obj, "index"):
            obj.index = index

    @classmethod
    def add_default_imports(cls, obj: xsd.Schema):
        """Add missing imports to the standard schemas if the namespace is
        declared and."""
        imp_namespaces = [imp.namespace for imp in obj.imports]
        xsi_ns = Namespace.XSI.uri
        if xsi_ns in obj.ns_map.values() and xsi_ns not in imp_namespaces:
            obj.imports.insert(0, xsd.Import(namespace=xsi_ns))

    @classmethod
    def reset_element_occurs(cls, obj: xsd.Schema):
        for element in obj.elements:
            element.min_occurs = None
            element.max_occurs = None



from typing import Type

from xsdata.formats.dataclass.parsers.handlers.native import XmlEventHandler
from xsdata.formats.dataclass.parsers.mixins import XmlHandler

try:
    from xsdata.formats.dataclass.parsers.handlers.lxml import LxmlEventHandler

    def default_handler() -> Type[XmlHandler]:
        return LxmlEventHandler

except ImportError:  # pragma: no cover

    def default_handler() -> Type[XmlHandler]:
        return XmlEventHandler


__all__ = [
    "LxmlEventHandler",
    "XmlEventHandler",
    "default_handler",
]



from .add_attribute_substitutions import AddAttributeSubstitutions
from .create_compound_fields import CreateCompoundFields
from .designate_class_packages import DesignateClassPackages
from .filter_classes import FilterClasses
from .flatten_attribute_groups import FlattenAttributeGroups
from .flatten_class_extensions import FlattenClassExtensions
from .merge_attributes import MergeAttributes
from .process_attributes_types import ProcessAttributeTypes
from .process_mixed_content_class import ProcessMixedContentClass
from .rename_duplicate_attributes import RenameDuplicateAttributes
from .rename_duplicate_classes import RenameDuplicateClasses
from .sanitize_attributes_default_value import SanitizeAttributesDefaultValue
from .sanitize_enumeration_class import SanitizeEnumerationClass
from .unnest_inner_classes import UnnestInnerClasses
from .update_attributes_effective_choice import UpdateAttributesEffectiveChoice
from .vacuum_inner_classes import VacuumInnerClasses
from .validate_attributes_overrides import ValidateAttributesOverrides

__all__ = [
    "AddAttributeSubstitutions",
    "CreateCompoundFields",
    "DesignateClassPackages",
    "FilterClasses",
    "FlattenAttributeGroups",
    "FlattenClassExtensions",
    "MergeAttributes",
    "ProcessAttributeTypes",
    "ProcessMixedContentClass",
    "RenameDuplicateAttributes",
    "RenameDuplicateClasses",
    "SanitizeAttributesDefaultValue",
    "SanitizeEnumerationClass",
    "UnnestInnerClasses",
    "UpdateAttributesEffectiveChoice",
    "VacuumInnerClasses",
    "ValidateAttributesOverrides",
]



import operator
import sys
from dataclasses import asdict
from dataclasses import dataclass
from dataclasses import field
from dataclasses import replace
from enum import IntEnum
from typing import Any
from typing import Dict
from typing import Iterator
from typing import List
from typing import Optional
from typing import Type

from xsdata.exceptions import CodeGenerationError
from xsdata.formats.converter import converter
from xsdata.formats.dataclass.models.elements import XmlType
from xsdata.models.enums import DataType
from xsdata.models.enums import Namespace
from xsdata.models.enums import Tag
from xsdata.models.mixins import ElementBase
from xsdata.utils import namespaces
from xsdata.utils import text

xml_type_map = {
    Tag.ANY: XmlType.WILDCARD,
    Tag.ANY_ATTRIBUTE: XmlType.ATTRIBUTES,
    Tag.ATTRIBUTE: XmlType.ATTRIBUTE,
    Tag.CHOICE: XmlType.ELEMENTS,
    Tag.ELEMENT: XmlType.ELEMENT,
}

SIMPLE_TYPES = (Tag.EXTENSION, Tag.LIST, Tag.SIMPLE_TYPE, Tag.UNION)
GLOBAL_TYPES = (Tag.ELEMENT, Tag.BINDING_OPERATION, Tag.BINDING_MESSAGE, Tag.MESSAGE)


@dataclass
class Restrictions:
    """
    Model representation of a dataclass field validation and type metadata.

    :param min_occurs:
    :param max_occurs:
    :param min_exclusive:
    :param min_inclusive:
    :param min_length:
    :param max_exclusive:
    :param max_inclusive:
    :param max_length:
    :param total_digits:
    :param fraction_digits:
    :param length:
    :param white_space:
    :param pattern:
    :param explicit_timezone:
    :param nillable:
    :param sequential:
    :param tokens:
    :param format:
    :param choice:
    """

    min_occurs: Optional[int] = field(default=None)
    max_occurs: Optional[int] = field(default=None)
    min_exclusive: Optional[str] = field(default=None)
    min_inclusive: Optional[str] = field(default=None)
    min_length: Optional[int] = field(default=None)
    max_exclusive: Optional[str] = field(default=None)
    max_inclusive: Optional[str] = field(default=None)
    max_length: Optional[int] = field(default=None)
    total_digits: Optional[int] = field(default=None)
    fraction_digits: Optional[int] = field(default=None)
    length: Optional[int] = field(default=None)
    white_space: Optional[str] = field(default=None)
    pattern: Optional[str] = field(default=None)
    explicit_timezone: Optional[str] = field(default=None)
    nillable: Optional[bool] = field(default=None)
    sequential: Optional[bool] = field(default=None)
    tokens: Optional[bool] = field(default=None)
    format: Optional[str] = field(default=None)
    choice: Optional[str] = field(default=None)

    @property
    def is_list(self) -> bool:
        """Return true if max occurs property is larger than one."""
        return self.max_occurs is not None and self.max_occurs > 1

    @property
    def is_optional(self) -> bool:
        """Return true if min occurs property equals zero."""
        return self.min_occurs == 0

    @property
    def is_prohibited(self) -> bool:
        return self.max_occurs == 0

    def merge(self, source: "Restrictions"):
        """Update properties from another instance."""

        self.update(source)

        min_occurs = source.min_occurs
        max_occurs = source.max_occurs
        is_list = max_occurs is not None and max_occurs > 1

        # Update the sequential flag if new value is true and restrictions indicate
        # the field was and still is a list.
        if source.sequential and (is_list or not self.is_list):
            self.sequential = source.sequential

        self.choice = source.choice or self.choice
        self.tokens = source.tokens or self.tokens
        self.format = source.format or self.format

        # Update min occurs if current value is None or the new value is more than one.
        if self.min_occurs is None or (min_occurs is not None and min_occurs != 1):
            self.min_occurs = min_occurs

        # Update max occurs if current value is None or the new value is more than one.
        if self.max_occurs is None or (max_occurs is not None and max_occurs != 1):
            self.max_occurs = max_occurs

    def update(self, source: "Restrictions"):
        keys = (
            "min_exclusive",
            "min_inclusive",
            "min_length",
            "max_exclusive",
            "max_inclusive",
            "max_length",
            "total_digits",
            "fraction_digits",
            "length",
            "white_space",
            "pattern",
            "explicit_timezone",
        )

        for key in keys:
            value = getattr(source, key)
            if value is not None:
                setattr(self, key, value)

    def asdict(self, types: Optional[List[Type]] = None) -> Dict:
        """
        Return the initialized only properties as a dictionary.

        Skip None or implied values, and optionally use the parent
        attribute types to convert relevant options.
        """

        result = {}
        sorted_types = converter.sort_types(types) if types else []

        if self.is_list:
            if self.min_occurs is not None and self.min_occurs > 0:
                result["min_occurs"] = self.min_occurs
            if self.max_occurs is not None and self.max_occurs < sys.maxsize:
                result["max_occurs"] = self.max_occurs
        elif self.min_occurs == self.max_occurs == 1 and not self.nillable:
            result["required"] = True

        for key, value in asdict(self).items():
            if value is None or key in ("choice", "min_occurs", "max_occurs"):
                continue

            if key.endswith("clusive") and types:
                value = converter.deserialize(value, sorted_types)

            result[key] = value

        return result

    def clone(self) -> "Restrictions":
        """Return a deep cloned instance."""
        return replace(self)

    @classmethod
    def from_element(cls, element: ElementBase) -> "Restrictions":
        """Static constructor from an xsd model."""
        return cls(**element.get_restrictions())


class AttrCategory(IntEnum):
    NATIVE = 0
    FORWARD = 1
    EXTERNAL = 2


@dataclass(unsafe_hash=True)
class AttrType:
    """
    Model representation for the typing information for fields and extensions.

    :param qname:
    :param alias:
    :param reference:
    :param native:
    :param forward:
    :param circular:
    :param substituted:
    """

    qname: str
    alias: Optional[str] = field(default=None, compare=False)
    reference: int = field(default=0, compare=False)
    native: bool = field(default=False)
    forward: bool = field(default=False)
    circular: bool = field(default=False)
    substituted: bool = field(default=False, compare=False)

    @property
    def datatype(self) -> Optional[DataType]:
        return DataType.from_qname(self.qname) if self.native else None

    @property
    def name(self) -> str:
        """Shortcut for qname local name."""
        return namespaces.local_name(self.qname)

    def is_dependency(self, allow_circular: bool) -> bool:
        """Return true if attribute is not a forward/circular references and
        it's not a native python time."""

        return not (
            self.forward or self.native or (not allow_circular and self.circular)
        )

    def clone(self) -> "AttrType":
        """Return a deep cloned instance."""
        return replace(self)


@dataclass
class Attr:
    """
    Model representation for a dataclass field.

    :param tag:
    :param name:
    :param local_name:
    :param index:
    :param default:
    :param fixed:
    :param mixed:
    :param types:
    :param choices:
    :param namespace:
    :param help:
    :param restrictions:
    """

    tag: str
    name: str = field(compare=False)
    local_name: str = field(init=False)
    index: int = field(compare=False, default_factory=int)
    default: Optional[str] = field(default=None, compare=False)
    fixed: bool = field(default=False, compare=False)
    mixed: bool = field(default=False, compare=False)
    types: List[AttrType] = field(default_factory=list, compare=False)
    choices: List["Attr"] = field(default_factory=list, compare=False)
    namespace: Optional[str] = field(default=None)
    help: Optional[str] = field(default=None, compare=False)
    restrictions: Restrictions = field(default_factory=Restrictions, compare=False)

    def __post_init__(self):
        self.local_name = self.name

    @property
    def is_attribute(self) -> bool:
        """Return whether this attribute is derived from an xs:attribute or
        xs:anyAttribute."""
        return self.tag in (Tag.ATTRIBUTE, Tag.ANY_ATTRIBUTE)

    @property
    def is_enumeration(self) -> bool:
        """Return whether this attribute is derived from an xs:enumeration."""
        return self.tag == Tag.ENUMERATION

    @property
    def is_dict(self) -> bool:
        """Return whether this attribute is a mapping of values."""
        return self.tag == Tag.ANY_ATTRIBUTE

    @property
    def is_factory(self) -> bool:
        """Return whether this attribute is a list of items or a mapping."""
        return self.is_list or self.is_dict or self.is_tokens

    @property
    def is_group(self) -> bool:
        """Return whether this attribute is derived from an xs:group or
        xs:attributeGroup."""
        return self.tag in (Tag.ATTRIBUTE_GROUP, Tag.GROUP)

    @property
    def is_list(self) -> bool:
        """Return whether this attribute is a list of values."""
        return self.restrictions.is_list

    @property
    def is_nameless(self) -> bool:
        """Return whether this attribute has a local name that will be used
        during parsing/serialization."""
        return self.tag not in (Tag.ATTRIBUTE, Tag.ELEMENT)

    @property
    def is_nillable(self) -> bool:
        return self.restrictions.nillable is True

    @property
    def is_optional(self) -> bool:
        """Return whether this attribute is not required."""
        return self.restrictions.is_optional

    @property
    def is_suffix(self) -> bool:
        """Return whether this attribute is not derived from an xs element with
        mode suffix."""
        return self.index == sys.maxsize

    @property
    def is_xsi_type(self) -> bool:
        """Return whether this attribute qualified name is equal to
        xsi:type."""
        return self.namespace == Namespace.XSI.uri and self.name == "type"

    @property
    def is_tokens(self) -> bool:
        """Return whether this attribute is a list of values."""
        return self.restrictions.tokens is True

    @property
    def is_wildcard(self) -> bool:
        """Return whether this attribute is derived from xs:anyAttribute or
        xs:any."""
        return self.tag in (Tag.ANY_ATTRIBUTE, Tag.ANY)

    @property
    def native_types(self) -> List[Type]:
        """Return a list of all builtin data types."""
        result = set()
        for tp in self.types:
            datatype = tp.datatype
            if datatype:
                result.add(datatype.type)

        return list(result)

    @property
    def user_types(self) -> Iterator[AttrType]:
        """Return an iterator of all the user defined types."""
        for tp in self.types:
            if not tp.native:
                yield tp

    @property
    def slug(self) -> str:
        return text.alnum(self.name)

    @property
    def xml_type(self) -> Optional[str]:
        """Return the xml node type this attribute is mapped to."""
        return xml_type_map.get(self.tag)

    def clone(self) -> "Attr":
        """Return a deep cloned instance."""
        return replace(
            self,
            types=[x.clone() for x in self.types],
            restrictions=self.restrictions.clone(),
        )


@dataclass(unsafe_hash=True)
class Extension:
    """
    Model representation of a dataclass base class.

    :param type:
    :param restrictions:
    """

    type: AttrType
    restrictions: Restrictions = field(hash=False)

    def clone(self) -> "Extension":
        """Return a deep cloned instance."""
        return replace(
            self,
            type=self.type.clone(),
            restrictions=self.restrictions.clone(),
        )


class Status(IntEnum):
    RAW = 0
    FLATTENING = 10
    FLATTENED = 11
    SANITIZING = 20
    SANITIZED = 21
    RESOLVING = 30
    RESOLVED = 31
    FINALIZING = 40
    FINALIZED = 41


@dataclass
class Class:
    """
    Model representation of a dataclass with fields, base/inner classes and
    additional metadata settings.

    :param qname:
    :param tag:
    :param location:
    :param mixed:
    :param abstract:
    :param nillable:
    :param local_type:
    :param status:
    :param container:
    :param package:
    :param module:
    :param namespace:
    :param help:
    :param meta_name:
    :param default:
    :param fixed:
    :param substitutions:
    :param extensions:
    :param attrs:
    :param inner:
    :param ns_map:
    """

    qname: str
    tag: str
    location: str
    mixed: bool = field(default=False)
    abstract: bool = field(default=False)
    nillable: bool = field(default=False)
    local_type: bool = field(default=False)
    status: Status = field(default=Status.RAW)
    container: Optional[str] = field(default=None)
    package: Optional[str] = field(default=None)
    module: Optional[str] = field(default=None)
    namespace: Optional[str] = field(default=None)
    help: Optional[str] = field(default=None)
    meta_name: Optional[str] = field(default=None)
    default: Any = field(default=None, compare=False)
    fixed: bool = field(default=False, compare=False)
    substitutions: List[str] = field(default_factory=list)
    extensions: List[Extension] = field(default_factory=list)
    attrs: List[Attr] = field(default_factory=list)
    inner: List["Class"] = field(default_factory=list)
    ns_map: Dict = field(default_factory=dict)

    @property
    def name(self) -> str:
        """Shortcut for qname local name."""
        return namespaces.local_name(self.qname)

    @property
    def slug(self) -> str:
        return text.alnum(self.name)

    @property
    def ref(self) -> int:
        return id(self)

    @property
    def target_namespace(self) -> Optional[str]:
        return namespaces.target_uri(self.qname)

    @property
    def has_suffix_attr(self) -> bool:
        """Return whether it includes a suffix attribute."""
        return any(attr.is_suffix for attr in self.attrs)

    @property
    def has_help_attr(self) -> bool:
        """Return whether it includes at least one attr with help content."""
        return any(attr.help and attr.help.strip() for attr in self.attrs)

    @property
    def is_complex(self) -> bool:
        """Return whether this instance is derived from an xs:element or
        xs:complexType."""
        return self.tag in (Tag.ELEMENT, Tag.COMPLEX_TYPE)

    @property
    def is_element(self) -> bool:
        """Return whether this instance is derived from an non abstract
        xs:element."""
        return self.tag == Tag.ELEMENT

    @property
    def is_enumeration(self) -> bool:
        """Return whether all attributes are derived from xs:enumeration."""
        return len(self.attrs) > 0 and all(attr.is_enumeration for attr in self.attrs)

    @property
    def is_global_type(self) -> bool:
        """Return whether this instance is a non-abstract element, wsdl binding
        class or a complex type without simple content."""
        return (not self.abstract and self.tag in GLOBAL_TYPES) or (
            self.tag == Tag.COMPLEX_TYPE and not self.is_simple_type
        )

    @property
    def is_group(self) -> bool:
        """Return whether this attribute is derived from an xs:group or
        xs:attributeGroup."""
        return self.tag in (Tag.ATTRIBUTE_GROUP, Tag.GROUP)

    @property
    def is_nillable(self) -> bool:
        """Return whether this class represents a nillable xml element."""
        return self.nillable or any(x.restrictions.nillable for x in self.extensions)

    @property
    def is_mixed(self) -> bool:
        """Return whether this class supports mixed content."""
        return self.mixed or any(x.mixed for x in self.attrs)

    @property
    def is_service(self) -> bool:
        """Return whether this instance is derived from wsdl:operation."""
        return self.tag == Tag.BINDING_OPERATION

    @property
    def is_simple_type(self) -> bool:
        """Return whether the class represents a simple text type."""
        return (
            len(self.attrs) == 1
            and self.attrs[0].tag in SIMPLE_TYPES
            and not self.extensions
        )

    @property
    def references(self) -> Iterator[int]:
        def all_refs():
            for ext in self.extensions:
                yield ext.type.reference

            for attr in self.attrs:
                for tp in attr.types:
                    yield tp.reference

                for choice in attr.choices:
                    for ctp in choice.types:
                        yield ctp.reference

            for inner in self.inner:
                yield from inner.references

        for ref in all_refs():
            if ref:
                yield ref

    @property
    def target_module(self) -> str:
        """Return the target module this class is assigned to."""
        if self.package and self.module:
            return f"{self.package}.{self.module}"

        if self.module:
            return self.module

        raise CodeGenerationError(
            f"Class `{self.name}` has not been assigned to a module yet!"
        )

    def clone(self) -> "Class":
        """Return a deep cloned instance."""
        inners = [inner.clone() for inner in self.inner]
        extensions = [extension.clone() for extension in self.extensions]
        attrs = [attr.clone() for attr in self.attrs]
        return replace(self, inner=inners, extensions=extensions, attrs=attrs)

    def dependencies(self, allow_circular: bool = False) -> Iterator[str]:
        """
        Return a set of dependencies for the given class.

        Collect:
            * base classes
            * attribute types
            * attribute choice types
            * recursively go through the inner classes
            * Ignore inner class references
            * Ignore native types.
        """

        types = {ext.type for ext in self.extensions}

        for attr in self.attrs:
            types.update(attr.types)
            types.update(tp for choice in attr.choices for tp in choice.types)

        for tp in types:
            if tp.is_dependency(allow_circular):
                yield tp.qname

        for inner in self.inner:
            yield from inner.dependencies(allow_circular)


@dataclass
class Import:
    """
    Model representation of a python import statement.

    :param qname:
    :param source:
    :param alias:
    """

    qname: str
    source: str
    alias: Optional[str] = field(default=None)

    @property
    def name(self) -> str:
        """Shortcut for qname local name."""
        return namespaces.local_name(self.qname)

    @property
    def slug(self) -> str:
        return text.alnum(self.name)


# Getters used all over the codegen process
get_location = operator.attrgetter("location")
get_name = operator.attrgetter("name")
get_qname = operator.attrgetter("qname")
get_tag = operator.attrgetter("tag")
get_restriction_choice = operator.attrgetter("restrictions.choice")
get_slug = operator.attrgetter("slug")
get_target_namespace = operator.attrgetter("target_namespace")
is_enumeration = operator.attrgetter("is_enumeration")
is_group = operator.attrgetter("is_group")



import sys
from collections import defaultdict
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import Set
from typing import Type

from xsdata.exceptions import XmlContextError
from xsdata.formats.bindings import T
from xsdata.formats.dataclass.compat import class_types
from xsdata.formats.dataclass.models.builders import XmlMetaBuilder
from xsdata.formats.dataclass.models.elements import XmlMeta
from xsdata.models.enums import DataType
from xsdata.utils.constants import return_input


class XmlContext:
    """
    The service provider for binding operations metadata.

    :param element_name_generator: Default element name generator
    :param attribute_name_generator: Default attribute name generator
    :param class_type: Default class type `dataclasses`
    """

    __slots__ = (
        "element_name_generator",
        "attribute_name_generator",
        "class_type",
        "cache",
        "xsi_cache",
        "sys_modules",
    )

    def __init__(
        self,
        element_name_generator: Callable = return_input,
        attribute_name_generator: Callable = return_input,
        class_type: str = "dataclasses",
    ):

        self.element_name_generator = element_name_generator
        self.attribute_name_generator = attribute_name_generator
        self.class_type = class_types.get_type(class_type)

        self.cache: Dict[Type, XmlMeta] = {}
        self.xsi_cache: Dict[str, List[Type]] = defaultdict(list)
        self.sys_modules = 0

    def reset(self):
        self.cache.clear()
        self.xsi_cache.clear()
        self.sys_modules = 0

    def fetch(
        self,
        clazz: Type,
        parent_ns: Optional[str] = None,
        xsi_type: Optional[str] = None,
    ) -> XmlMeta:
        """
        Fetch the model metadata of the given dataclass type, namespace and xsi
        type.

        :param clazz: The requested dataclass type
        :param parent_ns: The inherited parent namespace
        :param xsi_type: if present it means that the given clazz is derived and the
            lookup procedure needs to check and match a dataclass model to the qualified
            name instead
        """
        meta = self.build(clazz, parent_ns)
        subclass = None
        if xsi_type and meta.target_qname != xsi_type:
            subclass = self.find_subclass(clazz, xsi_type)

        return self.build(subclass, parent_ns) if subclass else meta

    def build_xsi_cache(self):
        """Index all imported dataclasses by their xsi:type qualified name."""

        if len(sys.modules) == self.sys_modules:
            return

        self.xsi_cache.clear()
        builder = XmlMetaBuilder(
            class_type=self.class_type,
            element_name_generator=self.element_name_generator,
            attribute_name_generator=self.attribute_name_generator,
        )
        for clazz in self.get_subclasses(object):
            if self.class_type.is_model(clazz):
                meta = builder.build_class_meta(clazz, None)

                if meta.target_qname:
                    self.xsi_cache[meta.target_qname].append(clazz)

        self.sys_modules = len(sys.modules)

    def find_types(self, qname: str) -> List[Type[T]]:
        """
        Find all classes that match the given xsi:type qname.

        - Ignores native schema types, xs:string, xs:float, xs:int, ...
        - Rebuild cache if new modules were imported since last run

        :param qname: Qualified name
        """
        if not DataType.from_qname(qname):
            self.build_xsi_cache()
            if qname in self.xsi_cache:
                return self.xsi_cache[qname]

        return []

    def find_type(self, qname: str) -> Optional[Type[T]]:
        """
        Return the most recently imported class that matches the given xsi:type
        qname.

        :param qname: Qualified name
        """
        types: List[Type] = self.find_types(qname)
        return types[-1] if types else None

    def find_type_by_fields(self, field_names: Set[str]) -> Optional[Type[T]]:
        """
        Find a dataclass from all the imported modules that matches the given
        list of field names.

        :param field_names: A unique list of field names
        """

        def get_field_diff(clazz: Type) -> int:
            meta = self.cache[clazz]
            local_names = {var.local_name for var in meta.get_all_vars()}
            return len(local_names - field_names)

        self.build_xsi_cache()
        choices = [
            (clazz, get_field_diff(clazz))
            for types in self.xsi_cache.values()
            for clazz in types
            if self.local_names_match(field_names, clazz)
        ]

        choices.sort(key=lambda x: (x[1], x[0].__name__))
        return choices[0][0] if len(choices) > 0 else None

    def find_subclass(self, clazz: Type, qname: str) -> Optional[Type]:
        """
        Compare all classes that match the given xsi:type qname and return the
        first one that is either a subclass or shares the same parent class as
        the original class.

        :param clazz: The search dataclass type
        :param qname: Qualified name
        """

        types: List[Type] = self.find_types(qname)
        for tp in types:

            # Why would an xml node with have an xsi:type that points
            # to parent class is beyond me but it happens, let's protect
            # against that scenario <node xsi:type="nodeAbstract" />
            if issubclass(clazz, tp):
                continue

            for tp_mro in tp.__mro__:
                if tp_mro is not object and tp_mro in clazz.__mro__:
                    return tp

        return None

    def build(
        self,
        clazz: Type,
        parent_ns: Optional[str] = None,
        globalns: Optional[Dict[str, Callable]] = None,
    ) -> XmlMeta:
        """
        Fetch from cache or build the binding metadata for the given class and
        parent namespace.

        :param clazz: A dataclass type
        :param parent_ns: The inherited parent namespace
        """

        if clazz not in self.cache:
            builder = XmlMetaBuilder(
                class_type=self.class_type,
                element_name_generator=self.element_name_generator,
                attribute_name_generator=self.attribute_name_generator,
                globalns=globalns,
            )
            self.cache[clazz] = builder.build(clazz, parent_ns)
        return self.cache[clazz]

    def build_recursive(self, clazz: Type, parent_ns: Optional[str] = None):
        """Build the binding metadata for the given class and all of its
        dependencies."""
        if clazz not in self.cache:
            meta = self.build(clazz, parent_ns)
            for var in meta.get_all_vars():
                types = var.element_types if var.elements else var.types
                for tp in types:
                    if self.class_type.is_model(tp):
                        self.build_recursive(tp, meta.namespace)

    def local_names_match(self, names: Set[str], clazz: Type) -> bool:
        try:
            meta = self.build(clazz)
            local_names = {var.local_name for var in meta.get_all_vars()}
            return not names.difference(local_names)
        except XmlContextError:
            return False

    @classmethod
    def is_derived(cls, obj: Any, clazz: Type) -> bool:
        """
        Return whether the given obj is derived from the given dataclass type.

        :param obj: A dataclass instance
        :param clazz: A dataclass type
        """

        if obj is None:
            return False

        if isinstance(obj, clazz):
            return True

        return any(x is not object and isinstance(obj, x) for x in clazz.__bases__)

    @classmethod
    def get_subclasses(cls, clazz: Type):
        try:
            for subclass in clazz.__subclasses__():
                yield from cls.get_subclasses(subclass)
                yield subclass
        except TypeError:
            pass



from collections import defaultdict
from typing import Dict
from typing import List
from typing import Optional

from xsdata.codegen.mixins import ContainerInterface
from xsdata.codegen.mixins import RelativeHandlerInterface
from xsdata.codegen.models import Attr
from xsdata.codegen.models import AttrType
from xsdata.codegen.models import Class
from xsdata.codegen.utils import ClassUtils
from xsdata.models.enums import Tag
from xsdata.utils import collections


class AddAttributeSubstitutions(RelativeHandlerInterface):
    """Apply substitution attributes to the given class recursively."""

    __slots__ = "substitutions"

    def __init__(self, container: ContainerInterface):
        super().__init__(container)
        self.substitutions: Optional[Dict[str, List[Attr]]] = None

    def process(self, target: Class):
        """
        Search and process attributes not derived from xs:enumeration or
        xs:any.

        Build the substitutions map if it's not initialized yet.
        """

        if self.substitutions is None:
            self.create_substitutions()

        for attr in list(target.attrs):
            if not (attr.is_enumeration or attr.is_wildcard):
                self.process_attribute(target, attr)

    def process_attribute(self, target: Class, attr: Attr):
        """
        Check if the given attribute matches any substitution class in order to
        clone it's attributes to the target class.

        The cloned attributes are placed below the attribute the are
        supposed to substitute.

        Guard against multiple substitutions in case of xs:groups.
        """
        index = target.attrs.index(attr)
        assert self.substitutions is not None

        for attr_type in attr.types:
            if attr_type.substituted:
                continue

            attr_type.substituted = True
            for substitution in self.substitutions.get(attr_type.qname, []):
                attr.restrictions.min_occurs = 0
                clone = ClassUtils.clone_attribute(substitution, attr.restrictions)

                pos = collections.find(target.attrs, clone)
                index = pos + 1 if pos > -1 else index
                target.attrs.insert(index, clone)

                self.process_attribute(target, clone)

    def create_substitutions(self):
        """Create reference attributes for all the classes substitutions and
        group them by their fully qualified name."""

        self.substitutions = defaultdict(list)
        for obj in self.container:
            for qname in obj.substitutions:
                attr = self.create_substitution(obj)
                self.substitutions[qname].append(attr)

    @classmethod
    def create_substitution(cls, source: Class) -> Attr:
        """Create an attribute with type that refers to the given source class
        and namespaced qualified name."""

        return Attr(
            name=source.name,
            types=[AttrType(qname=source.qname)],
            tag=Tag.ELEMENT,
            namespace=source.namespace,
        )



from typing import Iterator
from unittest import TestCase

from xsdata.models.xsd import MaxExclusive
from xsdata.models.xsd import MinExclusive
from xsdata.models.xsd import MinInclusive
from xsdata.models.xsd import MinLength
from xsdata.models.xsd import Restriction
from xsdata.models.xsd import SimpleType
from xsdata.models.xsd import Union


class UnionTests(TestCase):
    def test_property_bases(self):
        obj = Union()
        self.assertIsInstance(obj.bases, Iterator)
        self.assertEqual([], list(obj.bases))

        obj.member_types = "foo bar   "
        self.assertEqual(["foo", "bar"], list(obj.bases))

    def test_property_is_property(self):
        obj = Union()
        self.assertTrue(obj.is_property)

    def test_property_real_name(self):
        obj = Union()
        self.assertEqual("value", obj.real_name)

    def test_property_attr_types(self):
        obj = Union()
        obj.member_types = "thug life"
        self.assertEqual(["thug", "life"], list(obj.attr_types))

        obj = Union(
            simple_types=[
                SimpleType(restriction=Restriction(base="foo")),
                SimpleType(restriction=Restriction(base="bar")),
            ]
        )

        self.assertEqual(["foo", "bar"], list(obj.attr_types))

    def test_get_restrictions(self):
        first = Restriction(
            min_exclusive=MinExclusive(value="1"), min_inclusive=MinInclusive(value="2")
        )
        second = Restriction(
            min_length=MinLength(value="3"), max_exclusive=MaxExclusive(value="4")
        )
        obj = Union(
            simple_types=[
                SimpleType(restriction=first),
                SimpleType(restriction=second),
            ]
        )

        expected = {
            "max_exclusive": "4",
            "min_exclusive": "1",
            "min_inclusive": "2",
            "min_length": "3",
        }
        self.assertEqual(expected, obj.get_restrictions())



from dataclasses import dataclass, field
from enum import Enum
from typing import List, Optional


class PostStatus(Enum):
    DRAFT = "draft"
    PUBLISHED = "published"


@dataclass
class Tags:
    tag: List[str] = field(
        default_factory=list,
        metadata={
            "name": "Tag",
            "type": "Element",
        }
    )


@dataclass
class Post:
    status: PostStatus = field(
        default=PostStatus.DRAFT,
        metadata={
            "type": "Attribute",
            "required": True,
        }
    )
    lang: str = field(
        default="en",
        metadata={
            "type": "Attribute",
            "namespace": "http://www.w3.org/XML/1998/namespace",
            "required": True,
        }
    )
    created_at: Optional[str] = field(
        default=None,
        metadata={
            "type": "Attribute",
            "required": True,
        }
    )
    author: Optional[str] = field(
        default=None,
        metadata={
            "type": "Attribute",
            "required": True,
        }
    )
    origin: Optional[str] = field(
        default=None,
        metadata={
            "name": "Origin",
            "type": "Element",
        }
    )
    source: Optional[str] = field(
        default=None,
        metadata={
            "name": "Source",
            "type": "Element",
        }
    )
    title: Optional[str] = field(
        default=None,
        metadata={
            "name": "Title",
            "type": "Element",
            "required": True,
        }
    )
    body: Optional[str] = field(
        default=None,
        metadata={
            "name": "Body",
            "type": "Element",
            "required": True,
        }
    )
    tags: Optional[Tags] = field(
        default=None,
        metadata={
            "name": "Tags",
            "type": "Element",
            "required": True,
        }
    )


@dataclass
class Blog:
    post: List[Post] = field(
        default_factory=list,
        metadata={
            "name": "Post",
            "type": "Element",
            "min_occurs": 1,
        }
    )



from unittest import TestCase

from xsdata.models.wsdl import Binding
from xsdata.models.wsdl import BindingOperation


class BindingTests(TestCase):
    def test_unique_operations(self):
        binding = Binding(
            type="foo",
            operations=[
                BindingOperation(name="bar"),
                BindingOperation(name="bar"),
                BindingOperation(name="bar"),
                BindingOperation(name="foo"),
            ],
        )

        operations = list(binding.unique_operations())
        self.assertEqual(2, len(operations))
        self.assertEqual(binding.operations[2], operations[0])
        self.assertEqual(binding.operations[3], operations[1])



from tests.fixtures.books import Books, BookForm
from xsdata.models.datatype import XmlDate

books = Books(
    book=[
        BookForm(
            id="bk001",
            author="Hightower, Kim",
            title="The First Book",
            genre="Fiction",
            price=44.95,
            pub_date=XmlDate(2000, 10, 1),
            review="An amazing story of nothing.",
        ),
        BookForm(
            id="bk002",
            author="Nagata, Suanne",
            title="Becoming Somebody",
            genre="Biography",
            price=33.95,
            pub_date=XmlDate(2001, 1, 10),
            review="A masterpiece of the fine art of gossiping.",
        ),
    ]
)
events = [
    ('start-ns', 'brk', 'urn:books'),
    ('start', '{urn:books}books', {}, {'brk': 'urn:books'}),
    ('start', 'book', {'id': 'bk001', 'lang': 'en'}, {'brk': 'urn:books'}),
    ('start', 'author', {}, {'brk': 'urn:books'}),
    ('end', 'author', 'Hightower, Kim', '\n    '),
    ('start', 'title', {}, {'brk': 'urn:books'}),
    ('end', 'title', 'The First Book', '\n    '),
    ('start', 'genre', {}, {'brk': 'urn:books'}),
    ('end', 'genre', 'Fiction', '\n    '),
    ('start', 'price', {}, {'brk': 'urn:books'}),
    ('end', 'price', '44.95', '\n    '),
    ('start', 'pub_date', {}, {'brk': 'urn:books'}),
    ('end', 'pub_date', '2000-10-01', '\n    '),
    ('start', 'review', {}, {'brk': 'urn:books'}),
    ('end', 'review', 'An amazing story of nothing.', '\n  '),
    ('end', 'book', '\n    ', '\n  '),
    ('start', 'book', {'id': 'bk002', 'lang': 'en'}, {'brk': 'urn:books'}),
    ('start', 'author', {}, {'brk': 'urn:books'}),
    ('end', 'author', 'Nagata, Suanne', '\n    '),
    ('start', 'title', {}, {'brk': 'urn:books'}),
    ('end', 'title', 'Becoming Somebody', '\n    '),
    ('start', 'genre', {}, {'brk': 'urn:books'}),
    ('end', 'genre', 'Biography', '\n    '),
    ('start', 'price', {}, {'brk': 'urn:books'}),
    ('end', 'price', '33.95', '\n    '),
    ('start', 'pub_date', {}, {'brk': 'urn:books'}),
    ('end', 'pub_date', '2001-01-10', '\n    '),
    ('start', 'review', {}, {'brk': 'urn:books'}),
    ('end', 'review', 'A masterpiece of the fine art of gossiping.', '\n  '),
    ('end', 'book', '\n    ', '\n'),
    ('end', '{urn:books}books', '\n  ', None)
]

events_default_ns = [
    ("start-ns", None, "urn:books"),
    ("start", "{urn:books}books", {}, {None: "urn:books"}),
    ("start-ns", None, ""),
    ("start", "book", {"id": "bk001", "lang": "en"}, {None: ""}),
    ("start", "author", {}, {None: ""}),
    ("end", "author", "Hightower, Kim", "\n    "),
    ("start", "title", {}, {None: ""}),
    ("end", "title", "The First Book", "\n    "),
    ("start", "genre", {}, {None: ""}),
    ("end", "genre", "Fiction", "\n    "),
    ("start", "price", {}, {None: ""}),
    ("end", "price", "44.95", "\n    "),
    ("start", "pub_date", {}, {None: ""}),
    ("end", "pub_date", "2000-10-01", "\n    "),
    ("start", "review", {}, {None: ""}),
    ("end", "review", "An amazing story of nothing.", "\n  "),
    ("end", "book", "\n    ", "\n  "),
    ("start-ns", None, ""),
    ("start", "book", {"id": "bk002", "lang": "en"}, {None: ""}),
    ("start", "author", {}, {None: ""}),
    ("end", "author", "Nagata, Suanne", "\n    "),
    ("start", "title", {}, {None: ""}),
    ("end", "title", "Becoming Somebody", "\n    "),
    ("start", "genre", {}, {None: ""}),
    ("end", "genre", "Biography", "\n    "),
    ("start", "price", {}, {None: ""}),
    ("end", "price", "33.95", "\n    "),
    ("start", "pub_date", {}, {None: ""}),
    ("end", "pub_date", "2001-01-10", "\n    "),
    ("start", "review", {}, {None: ""}),
    ("end", "review", "A masterpiece of the fine art of gossiping.", "\n  "),
    ("end", "book", "\n    ", "\n"),
    ("end", "{urn:books}books", "\n  ", None),
]



import sys
from typing import Iterator
from unittest import mock

from xsdata.codegen.mappers.dtd import DtdMapper
from xsdata.codegen.models import Class
from xsdata.codegen.models import Restrictions
from xsdata.models.dtd import DtdAttributeDefault
from xsdata.models.dtd import DtdAttributeType
from xsdata.models.dtd import DtdContentOccur
from xsdata.models.dtd import DtdContentType
from xsdata.models.dtd import DtdElementType
from xsdata.models.enums import DataType
from xsdata.models.enums import Namespace
from xsdata.models.enums import Tag
from xsdata.utils.testing import AttrFactory
from xsdata.utils.testing import AttrTypeFactory
from xsdata.utils.testing import ClassFactory
from xsdata.utils.testing import DtdAttributeFactory
from xsdata.utils.testing import DtdContentFactory
from xsdata.utils.testing import DtdElementFactory
from xsdata.utils.testing import DtdFactory
from xsdata.utils.testing import FactoryTestCase


class DtdMapperTests(FactoryTestCase):
    def test_map(self):
        dtd = DtdFactory.root(2, location="tests.dtd")
        result = DtdMapper.map(dtd)

        self.assertIsInstance(result, Iterator)
        for cls in result:
            self.assertIsInstance(cls, Class)
            self.assertEqual(cls.location, dtd.location)

    @mock.patch.object(DtdMapper, "build_attributes")
    @mock.patch.object(DtdMapper, "build_elements")
    def test_build_class(self, mock_build_elements, mock_build_attributes):
        location = "tests.dtd"
        element = DtdElementFactory.create(
            name="root", prefix="ns", ns_map={"ns": "xsdata"}
        )
        actual = DtdMapper.build_class(element, "tests.dtd")
        expected = ClassFactory.create(
            qname="{xsdata}root",
            ns_map=element.ns_map,
            tag=Tag.ELEMENT,
            location=location,
            module=None,
        )

        self.assertEqual(expected, actual)

        mock_build_attributes.assert_called_once_with(actual, element)
        mock_build_elements.assert_called_once_with(actual, element)

    @mock.patch.object(DtdMapper, "build_attribute")
    def test_build_attributes(self, mock_build_attribute):
        attributes = DtdAttributeFactory.list(2)
        element = DtdElementFactory.create(attributes=attributes)
        target = ClassFactory.create()

        DtdMapper.build_attributes(target, element)

        mock_build_attribute.assert_has_calls(
            [
                mock.call(target, attributes[0]),
                mock.call(target, attributes[1]),
            ]
        )

    def test_build_attribute(self):
        target = ClassFactory.create()
        attribute = DtdAttributeFactory.create(
            name="lang", prefix="xml", default=DtdAttributeDefault.IMPLIED
        )
        DtdMapper.build_attribute(target, attribute)

        self.assertEqual(1, len(target.attrs))
        attr = target.attrs[0]

        self.assertEqual(0, attr.index)
        self.assertEqual(attribute.name, attr.name)
        self.assertEqual(Tag.ATTRIBUTE, attr.tag)
        self.assertEqual(Namespace.XML.uri, attr.namespace)
        self.assertEqual(1, len(attr.types))
        self.assertEqual(AttrTypeFactory.native(DataType.STRING), attr.types[0])
        self.assertEqual(1, attr.restrictions.max_occurs)
        self.assertEqual(0, attr.restrictions.min_occurs)

    def test_build_attribute_type_with_enumeration(self):
        target = ClassFactory.create()
        attribute = DtdAttributeFactory.create(
            type=DtdAttributeType.ENUMERATION, values=["a", "b", "c"]
        )
        actual = DtdMapper.build_attribute_type(target, attribute)

        self.assertTrue(actual.forward)
        self.assertEqual(attribute.name, actual.qname)

        self.assertEqual(1, len(target.inner))
        self.assertTrue(target.inner[0].is_enumeration)

        self.assertEqual(["a", "b", "c"], [x.name for x in target.inner[0].attrs])
        self.assertEqual(["a", "b", "c"], [x.default for x in target.inner[0].attrs])

        for attr in target.inner[0].attrs:
            self.assertEqual(Tag.ENUMERATION, attr.tag)
            self.assertTrue(attr.fixed)
            self.assertEqual([AttrTypeFactory.native(DataType.STRING)], attr.types)

    def test_build_attribute_restrictions_with_default_required(self):
        attr = AttrFactory.create()
        DtdMapper.build_attribute_restrictions(attr, DtdAttributeDefault.REQUIRED, "")
        self.assertEqual(1, attr.restrictions.max_occurs)
        self.assertEqual(1, attr.restrictions.min_occurs)
        self.assertFalse(attr.fixed)
        self.assertIsNone(attr.default)

    def test_build_attribute_restrictions_with_default_implied(self):
        attr = AttrFactory.create()
        DtdMapper.build_attribute_restrictions(attr, DtdAttributeDefault.IMPLIED, "")
        self.assertEqual(1, attr.restrictions.max_occurs)
        self.assertEqual(0, attr.restrictions.min_occurs)
        self.assertFalse(attr.fixed)
        self.assertIsNone(attr.default)

    def test_build_attribute_restrictions_with_default_fixed(self):
        attr = AttrFactory.create()
        DtdMapper.build_attribute_restrictions(attr, DtdAttributeDefault.FIXED, "abc")
        self.assertEqual(1, attr.restrictions.max_occurs)
        self.assertEqual(1, attr.restrictions.min_occurs)
        self.assertTrue(attr.fixed)
        self.assertEqual("abc", attr.default)

    def test_build_attribute_restrictions_with_default_none_and_no_value(self):
        attr = AttrFactory.create()
        DtdMapper.build_attribute_restrictions(attr, DtdAttributeDefault.NONE, None)
        self.assertEqual(1, attr.restrictions.max_occurs)
        self.assertEqual(0, attr.restrictions.min_occurs)
        self.assertFalse(attr.fixed)
        self.assertIsNone(attr.default)

    def test_build_attribute_restrictions_with_default_none_and_default_value(self):
        attr = AttrFactory.create()
        DtdMapper.build_attribute_restrictions(attr, DtdAttributeDefault.NONE, "abc")
        self.assertEqual(1, attr.restrictions.max_occurs)
        self.assertEqual(1, attr.restrictions.min_occurs)
        self.assertFalse(attr.fixed)
        self.assertEqual("abc", attr.default)

    def test_build_elements_with_mixed_element_type(self):
        target = ClassFactory.create(tag=Tag.ELEMENT)
        element = DtdElementFactory.create(type=DtdElementType.MIXED)

        DtdMapper.build_elements(target, element)
        self.assertEqual(0, len(target.attrs))
        self.assertEqual(1, len(target.extensions))

        self.assertEqual(Tag.COMPLEX_TYPE, target.tag)
        self.assertEqual(str(DataType.STRING), target.extensions[0].type.qname)
        self.assertTrue(target.extensions[0].type.native)

    def test_build_elements_with_any_element_type(self):
        target = ClassFactory.create()
        element = DtdElementFactory.create(type=DtdElementType.ANY)

        DtdMapper.build_elements(target, element)
        self.assertEqual(0, len(target.attrs))
        self.assertEqual(1, len(target.extensions))

        self.assertEqual(str(DataType.ANY_TYPE), target.extensions[0].type.qname)
        self.assertTrue(target.extensions[0].type.native)

        self.assertFalse(target.mixed)

    def test_build_elements_with_undefined_element_type(self):
        target = ClassFactory.create()
        element = DtdElementFactory.create(type=DtdElementType.UNDEFINED)

        DtdMapper.build_elements(target, element)
        self.assertEqual(0, len(target.attrs))
        self.assertEqual(0, len(target.extensions))
        self.assertFalse(target.mixed)

    def test_build_elements_with_empty_element_type(self):
        target = ClassFactory.create()
        element = DtdElementFactory.create(type=DtdElementType.EMPTY)

        DtdMapper.build_elements(target, element)
        self.assertEqual(0, len(target.attrs))
        self.assertEqual(0, len(target.extensions))
        self.assertFalse(target.mixed)

    def test_build_elements_with_element_element_type(self):
        target = ClassFactory.create()
        element = DtdElementFactory.create(
            type=DtdElementType.ELEMENT, content=DtdContentFactory.create(name="child")
        )

        DtdMapper.build_elements(target, element)
        self.assertEqual(1, len(target.attrs))
        self.assertEqual(0, len(target.extensions))
        self.assertFalse(target.mixed)

    def test_build_content_with_type_element(self):
        content = DtdContentFactory.create(
            type=DtdContentType.ELEMENT, occur=DtdContentOccur.ONCE
        )
        target = ClassFactory.create()
        DtdMapper.build_content(target, content, nillable=True)
        self.assertEqual(1, len(target.attrs))
        self.assertEqual(content.name, target.attrs[0].name)
        self.assertTrue(target.attrs[0].restrictions.nillable)
        self.assertEqual(1, target.attrs[0].restrictions.max_occurs)
        self.assertEqual(1, target.attrs[0].restrictions.min_occurs)

    def test_build_content_type_seq(self):
        content = DtdContentFactory.create(
            type=DtdContentType.SEQ,
            left=DtdContentFactory.create(type=DtdContentType.ELEMENT),
            right=DtdContentFactory.create(type=DtdContentType.ELEMENT),
        )
        target = ClassFactory.create()
        DtdMapper.build_content(target, content, nillable=True)
        self.assertEqual(2, len(target.attrs))
        self.assertTrue(target.attrs[0].restrictions.nillable)
        self.assertTrue(target.attrs[1].restrictions.nillable)

    def test_build_content_type_or(self):
        content = DtdContentFactory.create(
            type=DtdContentType.OR,
            left=DtdContentFactory.create(type=DtdContentType.ELEMENT),
            right=DtdContentFactory.create(type=DtdContentType.ELEMENT),
        )
        target = ClassFactory.create()
        DtdMapper.build_content(target, content, nillable=True)
        self.assertEqual(2, len(target.attrs))
        for attr in target.attrs:
            self.assertEqual(str(id(content)), attr.restrictions.choice)
            self.assertEqual(0, attr.restrictions.min_occurs)

    def test_build_content_type_or_nested(self):
        content = DtdContentFactory.create(
            type=DtdContentType.OR,
            left=DtdContentFactory.create(type=DtdContentType.ELEMENT),
            right=DtdContentFactory.create(type=DtdContentType.ELEMENT),
        )
        target = ClassFactory.create()
        DtdMapper.build_content(target, content, min_occurs=1, choice="abc")
        self.assertEqual(2, len(target.attrs))
        for attr in target.attrs:
            self.assertEqual("abc", attr.restrictions.choice)
            self.assertEqual(1, attr.restrictions.min_occurs)

    def test_build_content_type_pcdata(self):
        content = DtdContentFactory.create(
            type=DtdContentType.PCDATA,
        )
        target = ClassFactory.create()
        DtdMapper.build_content(target, content)
        self.assertEqual(0, len(target.attrs))

    @mock.patch.object(DtdMapper, "build_content")
    def test_build_content_tree(self, mock_build_content):
        center = DtdContentFactory.create()
        target = ClassFactory.create()

        DtdMapper.build_content_tree(target, center, nillable=True)

        center.left = DtdContentFactory.create()
        center.right = DtdContentFactory.create()
        DtdMapper.build_content_tree(target, center, nillable=False)

        mock_build_content.assert_has_calls(
            [
                mock.call(target, center.left, nillable=False),
                mock.call(target, center.right, nillable=False),
            ]
        )

    def test_build_restrictions(self):
        result = DtdMapper.build_restrictions(DtdContentOccur.ONCE)
        self.assertEqual(1, result.min_occurs)
        self.assertEqual(1, result.max_occurs)

        result = DtdMapper.build_restrictions(DtdContentOccur.OPT)
        self.assertEqual(0, result.min_occurs)
        self.assertEqual(1, result.max_occurs)

        result = DtdMapper.build_restrictions(DtdContentOccur.MULT)
        self.assertEqual(0, result.min_occurs)
        self.assertEqual(sys.maxsize, result.max_occurs)

        result = DtdMapper.build_restrictions(DtdContentOccur.PLUS)
        self.assertEqual(1, result.min_occurs)
        self.assertEqual(sys.maxsize, result.max_occurs)

        result = DtdMapper.build_restrictions(DtdContentOccur.PLUS, nillable=True)
        self.assertEqual(1, result.min_occurs)
        self.assertEqual(sys.maxsize, result.max_occurs)
        self.assertTrue(result.nillable)

        result = DtdMapper.build_restrictions(DtdContentOccur.PLUS, min_occurs=0)
        self.assertEqual(0, result.min_occurs)

    def test_build_element(self):
        target = ClassFactory.create()
        restrictions = Restrictions()
        names = ["firs", "second", "third"]

        for name in names:
            DtdMapper.build_element(target, name, restrictions)

        self.assertEqual(len(names), len(target.attrs))

        for index, attr in enumerate(target.attrs):
            self.assertEqual(names[index], attr.name)
            self.assertEqual(index, attr.index)
            self.assertEqual(Tag.ELEMENT, attr.tag)
            self.assertEqual(restrictions, attr.restrictions)
            self.assertIsNot(restrictions, attr.restrictions)

            self.assertEqual(1, len(attr.types))
            self.assertFalse(attr.types[0].native)
            self.assertEqual(names[index], attr.types[0].qname)



import datetime
import operator
import re
from collections import UserString
from typing import Any
from typing import Callable
from typing import Dict
from typing import NamedTuple
from typing import Optional
from typing import Union

from xsdata.utils.dates import calculate_offset
from xsdata.utils.dates import calculate_timezone
from xsdata.utils.dates import format_date
from xsdata.utils.dates import format_offset
from xsdata.utils.dates import format_time
from xsdata.utils.dates import parse_date_args
from xsdata.utils.dates import validate_date
from xsdata.utils.dates import validate_time

xml_duration_re = re.compile(
    r"^([-]?)P"
    r"(?:(\d+)Y)?(?:(\d+)M)?(?:(\d+)D)?"
    r"(?:T(?:(\d+)H)?(?:(\d+)M)?(?:(\d+(.\d+)?)S)?)?$"
)

DS_YEAR = 31556926.0
DS_MONTH = 2629743
DS_DAY = 86400
DS_HOUR = 3600
DS_MINUTE = 60
DS_FRACTIONAL_SECOND = 0.000000001
DS_OFFSET = -60


class DateFormat:
    DATE = "%Y-%m-%d%z"
    TIME = "%H:%M:%S%z"
    DATE_TIME = "%Y-%m-%dT%H:%M:%S%z"
    G_DAY = "---%d%z"
    G_MONTH = "--%m%z"
    G_MONTH_DAY = "--%m-%d%z"
    G_YEAR = "%Y%z"
    G_YEAR_MONTH = "%Y-%m%z"


class XmlDate(NamedTuple):
    """
    Concrete xs:date builtin type.

    Represents iso 8601 date format [-]CCYY-MM-DD[Z|(+|-)hh:mm] with
    rich comparisons and hashing.

    :param year: Any signed integer, eg (0, -535, 2020)
    :param month: Unsigned integer between 1-12
    :param day: Unsigned integer between 1-31
    :param offset: Signed integer representing timezone offset in minutes
    """

    year: int
    month: int
    day: int
    offset: Optional[int] = None

    def replace(
        self,
        year: Optional[int] = None,
        month: Optional[int] = None,
        day: Optional[int] = None,
        offset: Optional[int] = True,
    ) -> "XmlDate":
        """Return a new instance replacing the specified fields with new
        values."""

        if year is None:
            year = self.year
        if month is None:
            month = self.month
        if day is None:
            day = self.day
        if offset is True:
            offset = self.offset

        return type(self)(year, month, day, offset)

    @classmethod
    def from_string(cls, string: str) -> "XmlDate":
        """Initialize from string with format ``%Y-%m-%dT%z``"""
        return cls(*parse_date_args(string, DateFormat.DATE))

    @classmethod
    def from_date(cls, obj: datetime.date) -> "XmlDate":
        """
        Initialize from :class:`datetime.date` instance.

        .. warning::

            date instances don't have timezone information!
        """
        return cls(obj.year, obj.month, obj.day)

    @classmethod
    def from_datetime(cls, obj: datetime.datetime) -> "XmlDate":
        """Initialize from :class:`datetime.datetime` instance."""
        return cls(obj.year, obj.month, obj.day, calculate_offset(obj))

    @classmethod
    def today(cls) -> "XmlDate":
        """Initialize from datetime.date.today()"""
        return cls.from_date(datetime.date.today())

    def to_date(self) -> datetime.date:
        """Return a :class:`datetime.date` instance."""
        return datetime.date(self.year, self.month, self.day)

    def to_datetime(self) -> datetime.datetime:
        """Return a :class:`datetime.datetime` instance."""
        tz_info = calculate_timezone(self.offset)
        return datetime.datetime(self.year, self.month, self.day, tzinfo=tz_info)

    def __str__(self) -> str:
        """
        Return the date formatted according to ISO 8601 for xml.

        Examples:
            - 2001-10-26
            - 2001-10-26+02:00
            - 2001-10-26Z
        """
        return format_date(self.year, self.month, self.day) + format_offset(self.offset)

    def __repr__(self) -> str:
        args = [self.year, self.month, self.day, self.offset]
        if args[-1] is None:
            del args[-1]

        return f"{self.__class__.__qualname__}({', '.join(map(str, args))})"


class XmlDateTime(NamedTuple):
    """
    Concrete xs:dateTime builtin type.

    Represents iso 8601 date time format [-]CCYY-MM-DDThh:mm:ss[Z|(+|-)hh:mm]
    with rich comparisons and hashing.

    :param year: Any signed integer, eg (0, -535, 2020)
    :param month: Unsigned integer between 1-12
    :param day: Unsigned integer between 1-31
    :param hour: Unsigned integer between 0-24
    :param minute: Unsigned integer between 0-59
    :param second: Unsigned integer between 0-59
    :param fractional_second: Unsigned integer between 0-999999999
    :param offset: Signed integer representing timezone offset in
        minutes
    """

    year: int
    month: int
    day: int
    hour: int
    minute: int
    second: int
    fractional_second: int = 0
    offset: Optional[int] = None

    @property
    def microsecond(self) -> int:
        return self.fractional_second // 1000

    @property
    def duration(self) -> float:
        if self.year < 0:
            negative = True
            year = -self.year
        else:
            negative = False
            year = self.year

        total = (
            year * DS_YEAR
            + self.month * DS_MONTH
            + self.day * DS_DAY
            + self.hour * DS_HOUR
            + self.minute * DS_MINUTE
            + self.second
            + self.fractional_second * DS_FRACTIONAL_SECOND
            + (self.offset or 0) * DS_OFFSET
        )
        return -total if negative else total

    @classmethod
    def from_string(cls, string: str) -> "XmlDateTime":
        """Initialize from string with format ``%Y-%m-%dT%H:%M:%S%z``"""
        (
            year,
            month,
            day,
            hour,
            minute,
            second,
            fractional_second,
            offset,
        ) = parse_date_args(string, DateFormat.DATE_TIME)
        validate_date(year, month, day)
        validate_time(hour, minute, second, fractional_second)

        return cls(year, month, day, hour, minute, second, fractional_second, offset)

    @classmethod
    def from_datetime(cls, obj: datetime.datetime) -> "XmlDateTime":
        """Initialize from :class:`datetime.datetime` instance."""
        return cls(
            obj.year,
            obj.month,
            obj.day,
            obj.hour,
            obj.minute,
            obj.second,
            obj.microsecond * 1000,
            calculate_offset(obj),
        )

    @classmethod
    def now(cls, tz: Optional[datetime.timezone] = None) -> "XmlDateTime":
        """Initialize from datetime.datetime.now()"""
        return cls.from_datetime(datetime.datetime.now(tz=tz))

    @classmethod
    def utcnow(cls) -> "XmlDateTime":
        """Initialize from datetime.datetime.utcnow()"""
        return cls.from_datetime(datetime.datetime.utcnow())

    def to_datetime(self) -> datetime.datetime:
        """Return a :class:`datetime.datetime` instance."""
        return datetime.datetime(
            self.year,
            self.month,
            self.day,
            self.hour,
            self.minute,
            self.second,
            self.microsecond,
            tzinfo=calculate_timezone(self.offset),
        )

    def replace(
        self,
        year: Optional[int] = None,
        month: Optional[int] = None,
        day: Optional[int] = None,
        hour: Optional[int] = None,
        minute: Optional[int] = None,
        second: Optional[int] = None,
        fractional_second: Optional[int] = None,
        offset: Optional[int] = True,
    ) -> "XmlDateTime":
        """Return a new instance replacing the specified fields with new
        values."""

        if year is None:
            year = self.year
        if month is None:
            month = self.month
        if day is None:
            day = self.day
        if hour is None:
            hour = self.hour
        if minute is None:
            minute = self.minute
        if second is None:
            second = self.second
        if fractional_second is None:
            fractional_second = self.fractional_second
        if offset is True:
            offset = self.offset

        return type(self)(
            year, month, day, hour, minute, second, fractional_second, offset
        )

    def __str__(self) -> str:
        """
        Return the datetime formatted according to ISO 8601 for xml.

        Examples:
            - 2001-10-26T21:32:52
            - 2001-10-26T21:32:52+02:00
            - 2001-10-26T19:32:52Z
            - 2001-10-26T19:32:52.126789
            - 2001-10-26T21:32:52.126
            - -2001-10-26T21:32:52.126Z
        """
        return "{}T{}{}".format(
            format_date(self.year, self.month, self.day),
            format_time(self.hour, self.minute, self.second, self.fractional_second),
            format_offset(self.offset),
        )

    def __repr__(self) -> str:
        args = tuple(self)
        if args[-1] is None:
            args = args[:-1]

            if args[-1] == 0:
                args = args[:-1]

        return f"{self.__class__.__qualname__}({', '.join(map(str, args))})"

    def __eq__(self, other: Any) -> bool:
        return cmp(self, other, operator.eq)

    def __ne__(self, other: Any) -> bool:
        return cmp(self, other, operator.ne)

    def __lt__(self, other: Any) -> bool:
        return cmp(self, other, operator.lt)

    def __le__(self, other: Any) -> bool:
        return cmp(self, other, operator.le)

    def __gt__(self, other: Any) -> bool:
        return cmp(self, other, operator.gt)

    def __ge__(self, other: Any) -> bool:
        return cmp(self, other, operator.ge)


class XmlTime(NamedTuple):
    """
    Concrete xs:time builtin type.

    Represents iso 8601 time format hh:mm:ss[Z|(+|-)hh:mm]
    with rich comparisons and hashing.

    :param hour: Unsigned integer between 0-24
    :param minute: Unsigned integer between 0-59
    :param second: Unsigned integer between 0-59
    :param fractional_second: Unsigned integer between 0-999999999
    :param offset: Signed integer representing timezone offset in minutes
    """

    hour: int
    minute: int
    second: int
    fractional_second: int = 0
    offset: Optional[int] = None

    @property
    def microsecond(self) -> int:
        return self.fractional_second // 1000

    @property
    def duration(self) -> float:
        return (
            self.hour * DS_HOUR
            + self.minute * DS_MINUTE
            + self.second
            + self.fractional_second * DS_FRACTIONAL_SECOND
            + (self.offset or 0) * DS_OFFSET
        )

    def replace(
        self,
        hour: Optional[int] = None,
        minute: Optional[int] = None,
        second: Optional[int] = None,
        fractional_second: Optional[int] = None,
        offset: Optional[int] = True,
    ) -> "XmlTime":
        """Return a new instance replacing the specified fields with new
        values."""

        if hour is None:
            hour = self.hour
        if minute is None:
            minute = self.minute
        if second is None:
            second = self.second
        if fractional_second is None:
            fractional_second = self.fractional_second
        if offset is True:
            offset = self.offset

        return type(self)(hour, minute, second, fractional_second, offset)

    @classmethod
    def from_string(cls, string: str) -> "XmlTime":
        """Initialize from string format ``%H:%M:%S%z``"""
        hour, minute, second, fractional_second, offset = parse_date_args(
            string, DateFormat.TIME
        )
        validate_time(hour, minute, second, fractional_second)
        return cls(hour, minute, second, fractional_second, offset)

    @classmethod
    def from_time(cls, obj: datetime.time) -> "XmlTime":
        """Initialize from :class:`datetime.time` instance."""
        return cls(
            obj.hour,
            obj.minute,
            obj.second,
            obj.microsecond * 1000,
            calculate_offset(obj),
        )

    @classmethod
    def now(cls, tz: Optional[datetime.timezone] = None) -> "XmlTime":
        """Initialize from datetime.datetime.now()"""
        return cls.from_time(datetime.datetime.now(tz=tz).time())

    @classmethod
    def utcnow(cls) -> "XmlTime":
        """Initialize from datetime.datetime.utcnow()"""
        return cls.from_time(datetime.datetime.utcnow().time())

    def to_time(self) -> datetime.time:
        """Return a :class:`datetime.time` instance."""
        return datetime.time(
            self.hour,
            self.minute,
            self.second,
            self.microsecond,
            tzinfo=calculate_timezone(self.offset),
        )

    def __str__(self) -> str:
        """
        Return the time formatted according to ISO 8601 for xml.

        Examples:
            - 21:32:52
            - 21:32:52+02:00,
            - 19:32:52Z
            - 21:32:52.126789
            - 21:32:52.126Z
        """
        return "{}{}".format(
            format_time(self.hour, self.minute, self.second, self.fractional_second),
            format_offset(self.offset),
        )

    def __repr__(self) -> str:
        args = list(self)
        if args[-1] is None:
            del args[-1]

        return f"{self.__class__.__qualname__}({', '.join(map(str, args))})"

    def __eq__(self, other: Any) -> bool:
        return cmp(self, other, operator.eq)

    def __ne__(self, other: Any) -> bool:
        return cmp(self, other, operator.ne)

    def __lt__(self, other: Any) -> bool:
        return cmp(self, other, operator.lt)

    def __le__(self, other: Any) -> bool:
        return cmp(self, other, operator.le)

    def __gt__(self, other: Any) -> bool:
        return cmp(self, other, operator.gt)

    def __ge__(self, other: Any) -> bool:
        return cmp(self, other, operator.ge)


DurationType = Union[XmlTime, XmlDateTime]


def cmp(a: DurationType, b: DurationType, op: Callable) -> bool:
    if isinstance(b, a.__class__):
        return op(a.duration, b.duration)

    return NotImplemented


class TimeInterval(NamedTuple):
    negative: bool
    years: Optional[int]
    months: Optional[int]
    days: Optional[int]
    hours: Optional[int]
    minutes: Optional[int]
    seconds: Optional[float]


class XmlDuration(UserString):
    """
    Concrete xs:duration builtin type.

    Represents iso 8601 duration format PnYnMnDTnHnMnS
    with rich comparisons and hashing.

    Format PnYnMnDTnHnMnS:
        - **P**: literal value that starts the expression
        - **nY**: the number of years followed by a literal Y
        - **nM**: the number of months followed by a literal M
        - **nD**: the number of days followed by a literal D
        - **T**: literal value that separates date and time parts
        - **nH**: the number of hours followed by a literal H
        - **nM**: the number of minutes followed by a literal M
        - **nS**: the number of seconds followed by a literal S

    :param value: String representation of a xs:duration, eg **P2Y6M5DT12H**
    """

    def __init__(self, value: str) -> None:
        super().__init__(value)
        self._interval = self._parse_interval(value)

    @property
    def years(self) -> Optional[int]:
        """Number of years in the interval."""
        return self._interval.years

    @property
    def months(self) -> Optional[int]:
        """Number of months in the interval."""
        return self._interval.months

    @property
    def days(self) -> Optional[int]:
        """Number of days in the interval."""
        return self._interval.days

    @property
    def hours(self) -> Optional[int]:
        """Number of hours in the interval."""
        return self._interval.hours

    @property
    def minutes(self) -> Optional[int]:
        """Number of minutes in the interval."""
        return self._interval.minutes

    @property
    def seconds(self) -> Optional[float]:
        """Number of seconds in the interval."""
        return self._interval.seconds

    @property
    def negative(self) -> bool:
        """Negative flag of the interval."""
        return self._interval.negative

    @classmethod
    def _parse_interval(cls, value: str) -> TimeInterval:
        if not isinstance(value, str):
            raise ValueError("Value must be string")

        if len(value) < 3 or value.endswith("T"):
            raise ValueError(f"Invalid format '{value}'")

        match = xml_duration_re.match(value)
        if not match:
            raise ValueError(f"Invalid format '{value}'")

        sign, years, months, days, hours, minutes, seconds, _ = match.groups()
        return TimeInterval(
            negative=sign == "-",
            years=int(years) if years else None,
            months=int(months) if months else None,
            days=int(days) if days else None,
            hours=int(hours) if hours else None,
            minutes=int(minutes) if minutes else None,
            seconds=float(seconds) if seconds else None,
        )

    def asdict(self) -> Dict:
        return self._interval._asdict()

    def __repr__(self) -> str:
        return f'{self.__class__.__qualname__}("{self.data}")'


class TimePeriod(NamedTuple):
    year: Optional[int]
    month: Optional[int]
    day: Optional[int]
    offset: Optional[int]


class XmlPeriod(UserString):
    """
    Concrete xs:gYear/Month/Day builtin type.

    Represents iso 8601 period formats with rich comparisons and hashing.

    Formats:
        - xs:gDay: **---%d%z**
        - xs:gMonth: **--%m%z**
        - xs:gYear: **%Y%z**
        - xs:gMonthDay: **--%m-%d%z**
        - xs:gYearMonth: **%Y-%m%z**

    :param value: String representation of a xs:period, eg **--11-01Z**
    """

    def __init__(self, value: str) -> None:
        value = value.strip()
        super().__init__(value)
        self._period = self._parse_period(value)

    @property
    def year(self) -> Optional[int]:
        """Period year."""
        return self._period.year

    @property
    def month(self) -> Optional[int]:
        """Period month."""
        return self._period.month

    @property
    def day(self) -> Optional[int]:
        """Period day."""
        return self._period.day

    @property
    def offset(self) -> Optional[int]:
        """Period timezone offset in minutes."""
        return self._period.offset

    @classmethod
    def _parse_period(cls, value: str) -> TimePeriod:
        year = month = day = offset = None
        if value.startswith("---"):
            day, offset = parse_date_args(value, DateFormat.G_DAY)
        elif value.startswith("--"):
            # Bogus format --MM--, --05---05:00
            if value[4:6] == "--":
                value = value[:4] + value[6:]

            if len(value) in (4, 5, 10):  # fixed lengths with/out timezone
                month, offset = parse_date_args(value, DateFormat.G_MONTH)
            else:
                month, day, offset = parse_date_args(value, DateFormat.G_MONTH_DAY)
        else:
            end = len(value)
            if value.find(":") > -1:  # offset
                end -= 6

            if value[:end].rfind("-") > 3:  # Minimum position for month sep
                year, month, offset = parse_date_args(value, DateFormat.G_YEAR_MONTH)
            else:
                year, offset = parse_date_args(value, DateFormat.G_YEAR)

        validate_date(0, month or 1, day or 1)

        return TimePeriod(year=year, month=month, day=day, offset=offset)

    def as_dict(self) -> Dict:
        """Return date units as dict."""
        return self._period._asdict()

    def __repr__(self) -> str:
        return f'{self.__class__.__qualname__}("{self.data}")'

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, XmlPeriod):
            return self._period == other._period

        return NotImplemented


class XmlHexBinary(bytes):
    """
    Subclass bytes to infer base16 format.

    This type can be used with xs:anyType fields that don't have a
    format property to specify the target output format.
    """


class XmlBase64Binary(bytes):
    """
    Subclass bytes to infer base64 format.

    This type can be used with xs:anyType fields that don't have a
    format property to specify the target output format.
    """



from typing import List

from xsdata.codegen.mixins import ContainerHandlerInterface
from xsdata.codegen.models import Class
from xsdata.logger import logger
from xsdata.models.config import ClassFilterStrategy


class FilterClasses(ContainerHandlerInterface):
    """Filter classes for code generation based on the configuration output
    filter strategy."""

    __slots__ = ()

    def run(self):
        classes = []
        filter_strategy = self.container.config.output.filter_strategy
        if filter_strategy == ClassFilterStrategy.ALL_GLOBALS:
            classes = self.filter_all_globals()
        elif filter_strategy == ClassFilterStrategy.REFERRED_GLOBALS:
            classes = self.filter_referred_globals()

        if classes:
            self.container.set(classes)
        elif filter_strategy != ClassFilterStrategy.ALL:
            logger.warning(
                "The filter strategy '%s' returned no classes,"
                " will generate all types.",
                filter_strategy.value,
            )

    def filter_all_globals(self) -> List[Class]:
        """Filter all globals and any referenced types."""
        occurs = set()
        for obj in self.container:
            if obj.is_global_type:
                occurs.add(obj.ref)
                occurs.update(obj.references)

        return [obj for obj in self.container if obj.ref in occurs]

    def filter_referred_globals(self) -> List[Class]:
        """Filter globals with any references."""
        occurs = set()
        for obj in self.container:
            if obj.is_global_type:
                references = list(obj.references)
                occurs.update(references)
                if references:
                    occurs.add(obj.ref)

        return [obj for obj in self.container if obj.ref in occurs]



from unittest import mock

from xsdata.codegen.container import ClassContainer
from xsdata.codegen.handlers import FlattenAttributeGroups
from xsdata.codegen.models import Attr
from xsdata.codegen.models import Status
from xsdata.codegen.utils import ClassUtils
from xsdata.exceptions import AnalyzerValueError
from xsdata.models.config import GeneratorConfig
from xsdata.models.enums import Tag
from xsdata.utils.testing import AttrFactory
from xsdata.utils.testing import ClassFactory
from xsdata.utils.testing import FactoryTestCase


class FlattenAttributeGroupsTests(FactoryTestCase):
    def setUp(self):
        super().setUp()

        container = ClassContainer(config=GeneratorConfig())
        self.processor = FlattenAttributeGroups(container=container)

    @mock.patch.object(Attr, "is_group", new_callable=mock.PropertyMock)
    @mock.patch.object(FlattenAttributeGroups, "process_attribute")
    def test_process(self, mock_process_attribute, mock_is_group):
        mock_is_group.side_effect = [
            True,
            False,
            True,
            True,
            False,
            False,
        ]
        target = ClassFactory.elements(2)

        self.processor.process(target)
        self.assertEqual(6, mock_is_group.call_count)

        mock_process_attribute.assert_has_calls(
            [
                mock.call(target, target.attrs[0]),
                mock.call(target, target.attrs[0]),
                mock.call(target, target.attrs[1]),
            ]
        )

    @mock.patch.object(ClassUtils, "copy_group_attributes")
    def test_process_attribute_with_group(self, mock_copy_group_attributes):
        complex_bar = ClassFactory.create(qname="bar", tag=Tag.COMPLEX_TYPE)
        group_bar = ClassFactory.create(qname="bar", tag=Tag.ATTRIBUTE_GROUP)
        group_attr = AttrFactory.attribute_group(name="bar")
        target = ClassFactory.create()
        target.attrs.append(group_attr)

        self.processor.container.add(complex_bar)
        self.processor.container.add(group_bar)
        self.processor.container.add(target)

        self.processor.process_attribute(target, group_attr)
        mock_copy_group_attributes.assert_called_once_with(
            group_bar, target, group_attr
        )

    @mock.patch.object(ClassUtils, "copy_group_attributes")
    def test_process_attribute_with_attribute_group(self, mock_copy_group_attributes):
        complex_bar = ClassFactory.create(qname="bar", tag=Tag.COMPLEX_TYPE)
        group_bar = ClassFactory.create(qname="bar", tag=Tag.ATTRIBUTE_GROUP)
        group_attr = AttrFactory.attribute_group(name="bar")
        target = ClassFactory.create()
        target.attrs.append(group_attr)

        self.processor.container.add(complex_bar)
        self.processor.container.add(group_bar)
        self.processor.container.add(target)

        self.processor.process_attribute(target, group_attr)
        mock_copy_group_attributes.assert_called_once_with(
            group_bar, target, group_attr
        )

    def test_process_attribute_with_circular_reference(self):
        group_attr = AttrFactory.attribute_group(name="bar")
        target = ClassFactory.create(qname="bar", tag=Tag.ATTRIBUTE_GROUP)
        target.attrs.append(group_attr)

        target.status = Status.FLATTENING
        self.processor.container.add(target)

        self.processor.process_attribute(target, group_attr)
        self.assertFalse(group_attr in target.attrs)

    def test_process_attribute_with_unknown_source(self):
        group_attr = AttrFactory.attribute_group(name="bar")
        target = ClassFactory.create()
        target.attrs.append(group_attr)

        with self.assertRaises(AnalyzerValueError) as cm:
            self.processor.process_attribute(target, group_attr)

        self.assertEqual("Group attribute not found: `bar`", str(cm.exception))



import copy
from dataclasses import make_dataclass
from unittest import mock

from tests.fixtures.books import Books
from tests.fixtures.models import AttrsType
from tests.fixtures.models import ExtendedListType
from tests.fixtures.models import ExtendedType
from tests.fixtures.models import FixedType
from tests.fixtures.models import NillableType
from tests.fixtures.models import Paragraph
from tests.fixtures.models import SequentialType
from tests.fixtures.models import TypeA
from tests.fixtures.models import TypeB
from tests.fixtures.models import TypeC
from xsdata.exceptions import ParserError
from xsdata.formats.dataclass.context import XmlContext
from xsdata.formats.dataclass.models.elements import XmlType
from xsdata.formats.dataclass.models.generics import AnyElement
from xsdata.formats.dataclass.models.generics import DerivedElement
from xsdata.formats.dataclass.parsers.config import ParserConfig
from xsdata.formats.dataclass.parsers.nodes import ElementNode
from xsdata.formats.dataclass.parsers.nodes import PrimitiveNode
from xsdata.formats.dataclass.parsers.nodes import SkipNode
from xsdata.formats.dataclass.parsers.nodes import StandardNode
from xsdata.formats.dataclass.parsers.nodes import UnionNode
from xsdata.formats.dataclass.parsers.nodes import WildcardNode
from xsdata.formats.dataclass.parsers.utils import ParserUtils
from xsdata.models.enums import DataType
from xsdata.models.enums import Namespace
from xsdata.models.enums import QNames
from xsdata.utils.testing import FactoryTestCase
from xsdata.utils.testing import XmlMetaFactory
from xsdata.utils.testing import XmlVarFactory


class ElementNodeTests(FactoryTestCase):
    def setUp(self) -> None:
        super().setUp()
        self.context = XmlContext()
        self.meta = XmlMetaFactory.create(clazz=TypeC, qname="foo", wildcards=[])
        self.node = ElementNode(
            position=0,
            meta=self.meta,
            context=self.context,
            config=ParserConfig(),
            attrs={},
            ns_map={},
        )

    def test_bind(self):
        node = ElementNode(
            position=0,
            meta=self.context.build(SequentialType),
            context=self.context,
            config=ParserConfig(),
            attrs={"a": "b", "a0": "0"},
            ns_map={"ns0": "xsdata"},
        )

        objects = [("x1", 1), ("x2", 2), ("x2", 3)]
        expected = SequentialType(a0="0", a1={"a": "b"}, x0=1, x1=[1], x2=[2, 3])

        self.assertTrue(node.bind("foo", "1", "tail", objects))
        self.assertEqual("foo", objects[-1][0])
        self.assertEqual(expected, objects[-1][1])

    def test_bind_nil_value(self):
        self.node.xsi_nil = True
        objects = []

        self.assertTrue(self.node.bind("foo", None, None, objects))
        self.assertEqual(("foo", None), objects[-1])

    def test_bind_nillable_type(self):
        self.node.meta = self.context.build(NillableType)
        self.node.xsi_nil = True

        objects = []
        self.assertTrue(self.node.bind("foo", None, None, objects))
        self.assertEqual(("foo", NillableType(None)), objects[-1])

    def test_bind_fixed_value(self):
        self.node.meta = self.context.build(FixedType)

        objects = []
        self.assertTrue(self.node.bind("foo", "not the fixed value", None, objects))
        self.assertEqual(("foo", FixedType()), objects[-1])

    def test_bind_with_derived_element(self):
        self.node.meta = self.context.build(TypeA)
        self.node.derived_factory = DerivedElement

        objects = []
        self.assertTrue(self.node.bind("foo", "2", None, objects))
        self.assertEqual("foo", objects[-1][0])
        self.assertEqual(DerivedElement("foo", TypeA(2)), objects[-1][1])

    def test_bind_with_wildcard_var(self):
        self.node.meta = self.context.build(ExtendedType)
        self.node.attrs = {"a": "b"}
        self.node.ns_map = {"ns0": "xsdata"}

        objects = [("a", "1"), ("b", "2")]
        expected = ExtendedType(
            a="1",
            wildcard=AnyElement(
                text="text",
                tail="tail",
                children=[AnyElement(qname="b", text="2")],
                attributes={"a": "b"},
            ),
        )

        self.assertTrue(self.node.bind("foo", "text", "tail", objects))
        self.assertEqual("foo", objects[-1][0])
        self.assertEqual(expected, objects[-1][1])

    def test_bind_with_mixed_flag_true(self):
        self.node.meta = self.context.build(TypeB)
        self.node.attrs = {"a": "b"}
        self.node.ns_map = {"ns0": "xsdata"}
        self.node.mixed = True

        objects = [("x", 1), ("y", "a")]
        self.assertTrue(self.node.bind("foo", "text", "   ", objects))
        self.assertEqual(1, len(objects))
        self.assertEqual(TypeB(x=1, y="a"), objects[-1][1])

        objects = [("x", 1), ("y", "a")]
        self.assertTrue(self.node.bind("foo", "text", " tail ", objects))
        self.assertEqual(2, len(objects))
        self.assertEqual(None, objects[-1][0])
        self.assertEqual(" tail ", objects[-1][1])

    def test_bind_with_mixed_content_var(self):
        self.node.meta = self.context.build(Paragraph)
        self.node.attrs = {"a": "b"}
        self.node.ns_map = {"ns0": "xsdata"}

        objects = [("a", 1)]
        expected = Paragraph(content=["text", AnyElement(qname="a", text="1"), "tail"])

        self.assertTrue(self.node.bind("foo", "text", "tail", objects))
        self.assertEqual("foo", objects[-1][0])
        self.assertEqual(expected, objects[-1][1])

    def test_bind_wild_text(self):
        self.node.meta = self.context.build(ExtendedType)
        var = self.node.meta.wildcards[0]

        params = {}
        self.node.bind_wild_text(params, var, None, None)
        self.assertEqual(0, len(params))

        params = {}
        self.node.bind_wild_text(params, var, "txt", "tail")
        expected = AnyElement(text="txt", tail="tail")
        self.assertEqual({"wildcard": expected}, params)

        self.node.attrs = {"a": "b"}
        self.node.ns_map = {"ns0": "a"}
        self.node.bind_wild_text(params, var, "txt", "tail")
        expected = AnyElement(
            text="txt", tail="tail", children=[expected], attributes=self.node.attrs
        )
        self.assertEqual({"wildcard": expected}, params)

        self.node.meta = self.context.build(ExtendedListType)
        var = self.node.meta.wildcards[0]

        params = {}
        self.node.bind_wild_text(params, var, "txt", "tail")
        self.assertEqual({"wildcard": ["txt", "tail"]}, params)

        self.node.bind_wild_text(params, var, None, "tail")
        self.assertEqual({"wildcard": ["txt", "tail", "tail"]}, params)

        self.node.bind_wild_text(params, var, "first", None)
        self.assertEqual({"wildcard": ["first", "txt", "tail", "tail"]}, params)

    def test_bind_attrs(self):
        self.node.meta = self.context.build(AttrsType)
        self.node.attrs = {
            "index": "0",
            "fixed": "will be ignored",
            "{what}ever": "qname",
            "extended": "attr",
        }

        params = {}
        self.node.bind_attrs(params)

        expected = {"attrs": {"extended": "attr", "{what}ever": "qname"}, "index": 0}
        self.assertEqual(expected, params)

    def test_bind_attrs_with_fail_on_unknown_attributes(self):
        self.node.meta = self.context.build(AttrsType)
        self.node.config.fail_on_unknown_attributes = True
        self.node.attrs = {
            "index": "0",
            "fixed": "will be ignored",
            "{what}ever": "qname",
            "extended": "attr",
        }

        params = {}
        self.node.bind_attrs(params)

        expected = {"attrs": {"extended": "attr", "{what}ever": "qname"}, "index": 0}
        self.assertEqual(expected, params)

    def test_bind_with_fail_on_unknown_attributes(self):
        self.node.meta = self.context.build(ExtendedType)
        self.node.config.fail_on_unknown_attributes = True
        self.node.attrs = {"a": "b"}

        objects = [("a", "1")]
        with self.assertRaises(ParserError) as cm:
            self.node.bind("foo", "text", "tail", objects)

        self.assertEqual("Unknown attribute ExtendedType:a", str(cm.exception))

    @mock.patch("xsdata.formats.dataclass.parsers.nodes.element.logger.warning")
    def test_bind_objects(self, mock_warning):
        self.node.meta = self.context.build(TypeC)

        objects = [("x", 1), ("x", 2), ("z", 3.0), ("fixed", "bar")]

        params = {}
        self.node.bind_objects(params, objects)
        self.assertEqual({"x": 1, "z": 3.0}, params)

        mock_warning.assert_called_once_with("Unassigned parsed object %s", "x")

    def test_bind_wild_var(self):
        self.node.meta = self.context.build(ExtendedType)

        params = {}
        objects = [("x", 1), ("x", 2), ("z", 3.0)]
        self.node.bind_objects(params, objects)
        expected = {
            "wildcard": AnyElement(
                children=[
                    AnyElement(qname="x", text="1"),
                    AnyElement(qname="x", text="2"),
                    AnyElement(qname="z", text="3.0"),
                ]
            )
        }
        self.assertEqual(expected, params)

    def test_bind_wild_list_var(self):
        self.node.meta = self.context.build(ExtendedListType)

        params = {}
        objects = [("x", 1), ("x", 2), ("z", 3.0)]
        self.node.bind_objects(params, objects)
        expected = {
            "wildcard": [
                AnyElement(qname="x", text="1"),
                AnyElement(qname="x", text="2"),
                AnyElement(qname="z", text="3.0"),
            ]
        }
        self.assertEqual(expected, params)

    def test_prepare_generic_value(self):
        var = XmlVarFactory.create(
            index=2,
            xml_type=XmlType.WILDCARD,
            qname="a",
            types=(object,),
            elements={"known": XmlVarFactory.create()},
        )

        actual = self.node.prepare_generic_value(None, 1, var)
        self.assertEqual(1, actual)

        actual = self.node.prepare_generic_value("a", 1, var)
        expected = AnyElement(qname="a", text="1")
        self.assertEqual(expected, actual)

        actual = self.node.prepare_generic_value("a", "foo", var)
        expected = AnyElement(qname="a", text="foo")
        self.assertEqual(expected, actual)

        fixture = make_dataclass("Fixture", [("content", str)])
        actual = self.node.prepare_generic_value("a", fixture("foo"), var)
        expected = DerivedElement(qname="a", value=fixture("foo"), type="Fixture")
        self.assertEqual(expected, actual)

        fixture = make_dataclass("Fixture", [("content", str)])
        actual = self.node.prepare_generic_value("known", fixture("foo"), var)
        self.assertEqual(fixture("foo"), actual)

        actual = self.node.prepare_generic_value("a", expected, var)
        self.assertIs(expected, actual)

        actual = self.node.prepare_generic_value("Fixture", fixture("foo"), var)
        expected = DerivedElement(qname="Fixture", value=fixture("foo"))
        self.assertEqual(expected, actual)

    def test_child(self):
        var = XmlVarFactory.create(xml_type=XmlType.ELEMENT, qname="a", types=(TypeC,))
        attrs = {"a": "b"}
        ns_map = {"ns0": "xsdata"}
        position = 1
        self.meta.elements[var.qname] = [var]

        actual = self.node.child("a", attrs, ns_map, position)
        self.assertIsInstance(actual, ElementNode)
        self.assertEqual(attrs, actual.attrs)
        self.assertEqual(ns_map, actual.ns_map)
        self.assertEqual(position, actual.position)

    def test_child_with_unique_element(self):
        single = XmlVarFactory.create(
            index=1, xml_type=XmlType.ELEMENT, qname="a", types=(TypeC,)
        )
        wildcard = XmlVarFactory.create(
            index=2, xml_type=XmlType.WILDCARD, qname="a", types=(object,)
        )
        self.meta.elements[single.qname] = [single]
        self.meta.wildcards.append(wildcard)

        attrs = {"a": "b"}
        ns_map = {"ns0": "xsdata"}
        position = 1

        actual = self.node.child("a", attrs, ns_map, position)
        self.assertIsInstance(actual, ElementNode)
        self.assertIn(single.index, self.node.assigned)

        actual = self.node.child("a", attrs, ns_map, position)
        self.assertIsInstance(actual, WildcardNode)
        self.assertNotIn(wildcard.index, self.node.assigned)

    @mock.patch.object(ElementNode, "build_node")
    def test_child_when_failed_to_build_next_node(self, mock_build_node):
        mock_build_node.return_value = None
        element = XmlVarFactory.create(xml_type=XmlType.ELEMENT, qname="a")
        wildcard = XmlVarFactory.create(xml_type=XmlType.WILDCARD, qname="a")

        self.meta.elements[element.qname] = [element]
        self.meta.wildcards.append(wildcard)

        with self.assertRaises(ParserError) as cm:
            self.node.child("a", {}, {}, 0)

        self.assertEqual("Unknown property foo:a", str(cm.exception))

        self.node.config.fail_on_unknown_properties = False

        actual = self.node.child("foobar", {}, {}, 0)
        self.assertIsInstance(actual, SkipNode)

    def test_build_node_with_dataclass_union_var(self):
        var = XmlVarFactory.create(
            xml_type=XmlType.ELEMENT,
            name="a",
            qname="a",
            types=(TypeC, TypeB),
        )
        attrs = {"a": "b"}
        ns_map = {"ns0": "xsdata"}
        actual = self.node.build_node(var, attrs, ns_map, 10)

        self.assertIsInstance(actual, UnionNode)
        self.assertEqual(10, actual.position)
        self.assertIs(var, actual.var)
        self.assertEqual(attrs, actual.attrs)
        self.assertEqual(ns_map, actual.ns_map)
        self.assertEqual(0, actual.level)
        self.assertEqual(0, len(actual.events))

    @mock.patch.object(ParserUtils, "xsi_type", return_value="foo")
    @mock.patch.object(XmlContext, "fetch")
    def test_build_node_with_dataclass_var(self, mock_ctx_fetch, mock_xsi_type):
        var = XmlVarFactory.create(
            xml_type=XmlType.ELEMENT,
            name="a",
            qname="a",
            types=(TypeC,),
            derived=True,
        )
        xsi_type = "foo"
        namespace = self.meta.namespace
        mock_ctx_fetch.return_value = self.meta
        mock_xsi_type.return_value = xsi_type

        attrs = {"a": "b"}
        ns_map = {"ns0": "xsdata"}
        actual = self.node.build_node(var, attrs, ns_map, 10)

        self.assertIsInstance(actual, ElementNode)
        self.assertEqual(10, actual.position)
        self.assertEqual(DerivedElement, actual.derived_factory)
        self.assertIs(mock_ctx_fetch.return_value, actual.meta)

        mock_xsi_type.assert_called_once_with(attrs, ns_map)
        mock_ctx_fetch.assert_called_once_with(var.clazz, namespace, xsi_type)

    @mock.patch.object(XmlContext, "fetch")
    def test_build_node_with_dataclass_var_validates_nillable(self, mock_ctx_fetch):
        var = XmlVarFactory.create(xml_type=XmlType.ELEMENT, qname="a", types=(TypeC,))
        ns_map = {}
        nillable_meta = copy.deepcopy(self.meta)
        nillable_meta.nillable = True
        mock_ctx_fetch.side_effect = [self.meta, self.meta, nillable_meta]
        attrs = {QNames.XSI_NIL: "false"}

        self.assertIsNotNone(self.node.build_node(var, attrs, ns_map, 10))

        attrs = {QNames.XSI_NIL: "true"}
        self.assertIsNotNone(self.node.build_node(var, attrs, ns_map, 10))

        attrs = {QNames.XSI_NIL: "false"}
        self.assertIsNone(self.node.build_node(var, attrs, ns_map, 10))

    def test_build_node_with_any_type_var_with_matching_xsi_type(self):
        var = XmlVarFactory.create(
            xml_type=XmlType.ELEMENT,
            name="a",
            qname="a",
            types=(object,),
            any_type=True,
        )
        attrs = {QNames.XSI_TYPE: "bk:books"}
        ns_map = {"bk": "urn:books"}
        actual = self.node.build_node(var, attrs, ns_map, 10)

        self.assertIsInstance(actual, ElementNode)
        self.assertEqual(10, actual.position)
        self.assertEqual(self.context.build(Books), actual.meta)
        self.assertEqual(attrs, actual.attrs)
        self.assertEqual(ns_map, actual.ns_map)
        self.assertFalse(actual.mixed)

    def test_build_node_with_any_type_var_with_datatype(self):
        var = XmlVarFactory.create(
            xml_type=XmlType.ELEMENT,
            name="a",
            qname="a",
            types=(object,),
            any_type=True,
        )
        attrs = {QNames.XSI_TYPE: "xs:hexBinary"}
        ns_map = {Namespace.XS.prefix: Namespace.XS.uri}
        actual = self.node.build_node(var, attrs, ns_map, 10)

        self.assertIsInstance(actual, StandardNode)
        self.assertEqual(ns_map, actual.ns_map)
        self.assertEqual(DataType.HEX_BINARY, actual.datatype)
        self.assertIsNone(actual.derived_factory)

    def test_build_node_with_any_type_var_with_no_matching_xsi_type(self):
        var = XmlVarFactory.create(
            xml_type=XmlType.ELEMENT,
            name="a",
            qname="a",
            types=(object,),
            any_type=True,
        )
        attrs = {QNames.XSI_TYPE: "noMatch"}
        actual = self.node.build_node(var, attrs, {}, 10)

        self.assertIsInstance(actual, WildcardNode)
        self.assertEqual(10, actual.position)
        self.assertEqual(var, actual.var)
        self.assertEqual(attrs, actual.attrs)
        self.assertEqual({}, actual.ns_map)

    def test_build_node_with_any_type_var_with_no_xsi_type(self):
        var = XmlVarFactory.create(
            xml_type=XmlType.ELEMENT,
            name="a",
            qname="a",
            types=(object,),
            any_type=True,
        )
        attrs = {}
        actual = self.node.build_node(var, attrs, {}, 10)

        self.assertIsInstance(actual, WildcardNode)
        self.assertEqual(10, actual.position)
        self.assertEqual(var, actual.var)
        self.assertEqual(attrs, actual.attrs)
        self.assertEqual({}, actual.ns_map)

    def test_build_node_with_wildcard_var(self):
        var = XmlVarFactory.create(xml_type=XmlType.WILDCARD, qname="a")

        actual = self.node.build_node(var, {}, {}, 10)

        self.assertIsInstance(actual, WildcardNode)
        self.assertEqual(10, actual.position)
        self.assertEqual(var, actual.var)

    def test_build_node_with_primitive_var(self):
        var = XmlVarFactory.create(
            xml_type=XmlType.TEXT, qname="a", types=(int,), default=100
        )
        attrs = {"a": "b"}
        ns_map = {"ns0": "xsdata"}
        actual = self.node.build_node(var, attrs, ns_map, 10)

        self.assertIsInstance(actual, PrimitiveNode)
        self.assertEqual(ns_map, actual.ns_map)
        self.assertEqual(var, actual.var)



from .balance import (
    Available,
    Balance,
    ConnectReserved,
    Pending,
    SourceTypes,
)

__all__ = [
    "Available",
    "Balance",
    "ConnectReserved",
    "Pending",
    "SourceTypes",
]



import math
from typing import Any
from xml.etree.ElementTree import QName
from xml.sax.saxutils import quoteattr


def update(obj: Any, **kwargs: Any):
    """Update an object from keyword arguments with dotted keys."""

    for key, value in kwargs.items():
        attrsetter(obj, key, value)


def attrsetter(obj: Any, attr: str, value: Any):
    names = attr.split(".")
    last = names.pop()
    for name in names:
        obj = getattr(obj, name)

    setattr(obj, last, value)


def literal_value(value: Any) -> str:
    if isinstance(value, str):
        return quoteattr(value)

    if isinstance(value, float):
        return str(value) if math.isfinite(value) else f'float("{value}")'

    if isinstance(value, QName):
        return f'QName("{value.text}")'

    return repr(value).replace("'", '"')



import pytest

from xsdata.models.datatype import XmlDate
from xsdata.models.datatype import XmlDateTime
from xsdata.models.datatype import XmlDuration
from xsdata.models.datatype import XmlTime


@pytest.mark.benchmark(disable_gc=True, group="converters", min_rounds=100000)
def test_xml_date_time(benchmark):
    benchmark(XmlDateTime.from_string, "2010-09-20T13:00:00.000+01:00")


@pytest.mark.benchmark(disable_gc=True, group="converters", min_rounds=100000)
def test_xml_time(benchmark):
    benchmark(XmlTime.from_string, "13:00:00.000+01:00")


@pytest.mark.benchmark(disable_gc=True, group="converters", min_rounds=100000)
def test_xml_date(benchmark):
    benchmark(XmlDate.from_string, "2010-09-20")


@pytest.mark.benchmark(disable_gc=True, group="converters", min_rounds=100000)
def test_xml_duration(benchmark):
    benchmark(XmlDuration, "P2Y6M5DT12H35M30.5S")



from unittest import mock

from xsdata.codegen.container import ClassContainer
from xsdata.codegen.handlers import RenameDuplicateClasses
from xsdata.models.config import GeneratorConfig
from xsdata.models.config import StructureStyle
from xsdata.models.enums import Tag
from xsdata.utils.testing import AttrFactory
from xsdata.utils.testing import AttrTypeFactory
from xsdata.utils.testing import ClassFactory
from xsdata.utils.testing import ExtensionFactory
from xsdata.utils.testing import FactoryTestCase


class RenameDuplicateClassesTests(FactoryTestCase):
    def setUp(self):
        super().setUp()

        self.container = ClassContainer(config=GeneratorConfig())
        self.processor = RenameDuplicateClasses(container=self.container)

    @mock.patch.object(RenameDuplicateClasses, "rename_classes")
    def test_run(self, mock_rename_classes):
        classes = [
            ClassFactory.create(qname="{foo}A"),
            ClassFactory.create(qname="{foo}a"),
            ClassFactory.create(qname="_a"),
            ClassFactory.create(qname="_b"),
            ClassFactory.create(qname="b", location="!@#$"),
        ]
        self.container.extend(classes)
        self.processor.run()

        mock_rename_classes.assert_has_calls(
            [
                mock.call(classes[:2], False),
                mock.call(classes[3:], False),
            ]
        )

    @mock.patch.object(RenameDuplicateClasses, "rename_classes")
    def test_run_with_single_package_structure(self, mock_rename_classes):
        classes = [
            ClassFactory.create(qname="{foo}a"),
            ClassFactory.create(qname="{bar}a"),
            ClassFactory.create(qname="a"),
        ]
        self.container.extend(classes)
        self.processor.run()

        mock_rename_classes.assert_called_once_with(classes, True)

    @mock.patch.object(RenameDuplicateClasses, "rename_classes")
    def test_run_with_single_location_source(self, mock_rename_classes):
        classes = [
            ClassFactory.create(qname="{foo}a"),
            ClassFactory.create(qname="{bar}a"),
            ClassFactory.create(qname="a"),
        ]

        self.container.config.output.structure_style = StructureStyle.SINGLE_PACKAGE
        self.container.extend(classes)
        self.processor.run()

        mock_rename_classes.assert_called_once_with(classes, True)

    @mock.patch.object(RenameDuplicateClasses, "rename_classes")
    def test_run_with_clusters_structure(self, mock_rename_classes):
        classes = [
            ClassFactory.create(qname="{foo}a"),
            ClassFactory.create(qname="{bar}a"),
            ClassFactory.create(qname="a"),
        ]
        self.container.config.output.structure_style = StructureStyle.CLUSTERS
        self.container.extend(classes)
        self.processor.run()

        mock_rename_classes.assert_called_once_with(classes, True)

    @mock.patch.object(RenameDuplicateClasses, "rename_class")
    def test_rename_classes(self, mock_rename_class):
        classes = [
            ClassFactory.create(qname="_a", tag=Tag.ELEMENT),
            ClassFactory.create(qname="_A", tag=Tag.ELEMENT),
            ClassFactory.create(qname="a", tag=Tag.COMPLEX_TYPE),
        ]
        self.processor.rename_classes(classes, False)
        self.processor.rename_classes(classes, True)

        mock_rename_class.assert_has_calls(
            [
                mock.call(classes[1], False),
                mock.call(classes[0], False),
                mock.call(classes[2], False),
                mock.call(classes[1], True),
                mock.call(classes[0], True),
                mock.call(classes[2], True),
            ]
        )

    @mock.patch.object(RenameDuplicateClasses, "rename_class")
    def test_rename_classes_protects_single_element(self, mock_rename_class):
        classes = [
            ClassFactory.create(qname="_a", tag=Tag.ELEMENT),
            ClassFactory.create(qname="a", tag=Tag.COMPLEX_TYPE),
        ]
        self.processor.rename_classes(classes, False)

        mock_rename_class.assert_called_once_with(classes[1], False)

    @mock.patch.object(RenameDuplicateClasses, "rename_class_dependencies")
    def test_rename_class(self, mock_rename_class_dependencies):
        target = ClassFactory.create(qname="{foo}_a")
        self.processor.container.add(target)
        self.processor.container.add(ClassFactory.create(qname="{foo}a_1"))
        self.processor.container.add(ClassFactory.create(qname="{foo}A_2"))
        self.processor.container.add(ClassFactory.create(qname="{bar}a_3"))
        self.processor.rename_class(target, False)

        self.assertEqual("{foo}_a_3", target.qname)
        self.assertEqual("_a", target.meta_name)

        mock_rename_class_dependencies.assert_has_calls(
            mock.call(item, id(target), "{foo}_a_3")
            for item in self.processor.container
        )

        self.assertEqual([target], self.container.data["{foo}_a_3"])
        self.assertEqual([], self.container.data["{foo}_a"])

    @mock.patch.object(RenameDuplicateClasses, "rename_class_dependencies")
    def test_rename_class_by_name(self, mock_rename_class_dependencies):
        target = ClassFactory.create(qname="{foo}_a")
        self.processor.container.add(target)
        self.processor.container.add(ClassFactory.create(qname="{bar}a_1"))
        self.processor.container.add(ClassFactory.create(qname="{thug}A_2"))
        self.processor.container.add(ClassFactory.create(qname="{bar}a_3"))
        self.processor.rename_class(target, True)

        self.assertEqual("{foo}_a_4", target.qname)
        self.assertEqual("_a", target.meta_name)

        mock_rename_class_dependencies.assert_has_calls(
            mock.call(item, id(target), "{foo}_a_4")
            for item in self.processor.container
        )

        self.assertEqual([target], self.container.data["{foo}_a_4"])
        self.assertEqual([], self.container.data["{foo}_a"])

    def test_rename_class_dependencies(self):
        attr_type = AttrTypeFactory.create(qname="{foo}bar", reference=1)

        target = ClassFactory.create(
            extensions=[
                ExtensionFactory.create(),
                ExtensionFactory.create(attr_type.clone()),
            ],
            attrs=[
                AttrFactory.create(),
                AttrFactory.create(types=[AttrTypeFactory.create(), attr_type.clone()]),
            ],
            inner=[
                ClassFactory.create(
                    extensions=[ExtensionFactory.create(attr_type.clone())],
                    attrs=[
                        AttrFactory.create(),
                        AttrFactory.create(
                            types=[AttrTypeFactory.create(), attr_type.clone()]
                        ),
                    ],
                )
            ],
        )

        self.processor.rename_class_dependencies(target, 1, "thug")
        dependencies = set(target.dependencies())
        self.assertNotIn("{foo}bar", dependencies)
        self.assertIn("thug", dependencies)

    def test_rename_attr_dependencies_with_default_enum(self):
        attr_type = AttrTypeFactory.create(qname="{foo}bar", reference=1)
        target = ClassFactory.create(
            attrs=[
                AttrFactory.create(
                    types=[attr_type],
                    default=f"@enum@{attr_type.qname}::member",
                ),
            ]
        )

        self.processor.rename_class_dependencies(target, 1, "thug")
        dependencies = set(target.dependencies())
        self.assertEqual("@enum@thug::member", target.attrs[0].default)
        self.assertNotIn("{foo}bar", dependencies)
        self.assertIn("thug", dependencies)

    def test_rename_attr_dependencies_with_choices(self):
        attr_type = AttrTypeFactory.create(qname="foo", reference=1)
        target = ClassFactory.create(
            attrs=[
                AttrFactory.create(
                    choices=[
                        AttrFactory.create(types=[attr_type.clone()]),
                    ]
                )
            ]
        )

        self.processor.rename_class_dependencies(target, 1, "bar")
        dependencies = set(target.dependencies())
        self.assertNotIn("foo", dependencies)
        self.assertIn("bar", dependencies)



from typing import List

from xsdata.codegen.mixins import ContainerHandlerInterface
from xsdata.codegen.models import Attr
from xsdata.codegen.models import Class
from xsdata.codegen.models import get_location
from xsdata.codegen.models import get_name
from xsdata.codegen.models import get_qname
from xsdata.models.config import StructureStyle
from xsdata.utils import collections
from xsdata.utils import namespaces
from xsdata.utils import text

REQUIRE_UNIQUE_NAMES = (StructureStyle.SINGLE_PACKAGE, StructureStyle.CLUSTERS)


class RenameDuplicateClasses(ContainerHandlerInterface):
    """Resolve class name conflicts depending on the output structure style."""

    __slots__ = ()

    def run(self):
        """Search for conflicts either by qualified name or local name
        depending on the configuration and start renaming classes and
        dependencies."""

        use_name = self.should_use_names()
        getter = get_name if use_name else get_qname
        groups = collections.group_by(self.container, lambda x: text.alnum(getter(x)))

        for classes in groups.values():
            if len(classes) > 1:
                self.rename_classes(classes, use_name)

    def should_use_names(self) -> bool:
        """
        Determine if we should be using names or qualified names to detect
        collisions.

        Strict unique names:
        - Single package
        - Clustered packages
        - All classes have the same source location.
        """
        return (
            self.container.config.output.structure_style in REQUIRE_UNIQUE_NAMES
            or len(set(map(get_location, self.container))) == 1
        )

    def rename_classes(self, classes: List[Class], use_name: bool):
        """
        Rename all the classes in the list.

        Protect classes derived from xs:element if there is only one in
        the list.
        """
        total_elements = sum(x.is_element for x in classes)
        for target in sorted(classes, key=get_name):
            if not target.is_element or total_elements > 1:
                self.rename_class(target, use_name)

    def rename_class(self, target: Class, use_name: bool):
        """Find the next available class identifier, save the original name in
        the class metadata and update the class qualified name and all classes
        that depend on the target class."""

        qname = target.qname
        namespace, name = namespaces.split_qname(target.qname)
        target.qname = self.next_qname(namespace, name, use_name)
        target.meta_name = name
        self.container.reset(target, qname)

        for item in self.container:
            self.rename_class_dependencies(item, id(target), target.qname)

    def next_qname(self, namespace: str, name: str, use_name: bool) -> str:
        """Append the next available index number for the given namespace and
        local name."""
        index = 0

        if use_name:
            reserved = {text.alnum(obj.name) for obj in self.container}
        else:
            reserved = {text.alnum(obj.qname) for obj in self.container}

        while True:
            index += 1
            new_name = f"{name}_{index}"
            qname = namespaces.build_qname(namespace, new_name)
            cmp = text.alnum(new_name if use_name else qname)

            if cmp not in reserved:
                return qname

    def rename_class_dependencies(self, target: Class, reference: int, replace: str):
        """Search and replace the old qualified attribute type name with the
        new one if it exists in the target class attributes, extensions and
        inner classes."""
        for attr in target.attrs:
            self.rename_attr_dependencies(attr, reference, replace)

        for ext in target.extensions:
            if ext.type.reference == reference:
                ext.type.qname = replace

        for inner in target.inner:
            self.rename_class_dependencies(inner, reference, replace)

    def rename_attr_dependencies(self, attr: Attr, reference: int, replace: str):
        """Search and replace the old qualified attribute type name with the
        new one in the attr types, choices and default value."""
        for attr_type in attr.types:
            if attr_type.reference == reference:
                attr_type.qname = replace

                if isinstance(attr.default, str) and attr.default.startswith("@enum@"):
                    members = text.suffix(attr.default, "::")
                    attr.default = f"@enum@{replace}::{members}"

        for choice in attr.choices:
            self.rename_attr_dependencies(choice, reference, replace)



from typing import Optional

from xsdata.codegen.mixins import RelativeHandlerInterface
from xsdata.codegen.models import Attr
from xsdata.codegen.models import Class
from xsdata.utils.namespaces import build_qname


class UnnestInnerClasses(RelativeHandlerInterface):
    """Unnest class processor."""

    __slots__ = ()

    def process(self, target: Class):
        """
        Promote enumeration classes to root classes.

        Candidates
            - Enumerations
            - All if config is enabled
        """
        for inner in list(target.inner):
            if inner.is_enumeration or self.container.config.output.unnest_classes:
                self.promote(target, inner)

    def promote(self, target: Class, inner: Class):
        target.inner.remove(inner)
        attr = self.find_forward_attr(target, inner.qname)
        if attr:
            clone = self.clone_class(inner, target.name)
            self.update_types(attr, inner.qname, clone.qname)
            self.container.add(clone)

    @classmethod
    def clone_class(cls, inner: Class, name: str) -> Class:
        clone = inner.clone()
        clone.local_type = True
        clone.qname = build_qname(inner.target_namespace, f"{name}_{inner.name}")
        return clone

    @classmethod
    def update_types(cls, attr: Attr, search: str, replace: str):
        for attr_type in attr.types:
            if attr_type.qname == search and attr_type.forward:
                attr_type.qname = replace
                attr_type.forward = False

    @classmethod
    def find_forward_attr(cls, target: Class, qname: str) -> Optional[Attr]:
        for attr in target.attrs:
            for attr_type in attr.types:
                if attr_type.forward and attr_type.qname == qname:
                    return attr

        return None



from pathlib import Path
from typing import Iterator
from typing import List
from typing import Optional

from jinja2 import Environment
from jinja2 import FileSystemLoader

from xsdata.codegen.models import Class
from xsdata.codegen.models import Import
from xsdata.codegen.resolver import DependenciesResolver
from xsdata.formats.dataclass.filters import Filters
from xsdata.formats.mixins import AbstractGenerator
from xsdata.formats.mixins import GeneratorResult
from xsdata.models.config import GeneratorConfig


class DataclassGenerator(AbstractGenerator):
    """Python dataclasses code generator."""

    __slots__ = ("env", "filters")

    def __init__(self, config: GeneratorConfig):
        """Override generator constructor to set templates directory and
        environment filters."""

        super().__init__(config)

        tpl_dir = Path(__file__).parent.joinpath("templates")
        self.env = Environment(loader=FileSystemLoader(str(tpl_dir)), autoescape=False)
        self.filters = self.init_filters(config)
        self.filters.register(self.env)

    def render(self, classes: List[Class]) -> Iterator[GeneratorResult]:
        """
        Return an iterator of the generated results.

        Group classes into modules and yield an output per module and
        per path __init__.py file.
        """
        packages = {obj.qname: obj.target_module for obj in classes}
        resolver = DependenciesResolver(packages=packages)

        # Generate packages
        for path, cluster in self.group_by_package(classes).items():
            module = ".".join(path.relative_to(Path.cwd()).parts)
            yield GeneratorResult(
                path=path.joinpath("__init__.py"),
                title="init",
                source=self.render_package(cluster, module),
            )
            yield from self.ensure_packages(path.parent)

        # Generate modules
        for path, cluster in self.group_by_module(classes).items():
            yield GeneratorResult(
                path=path.with_suffix(".py"),
                title=cluster[0].target_module,
                source=self.render_module(resolver, cluster),
            )

    def render_package(self, classes: List[Class], module: str) -> str:
        """Render the source code for the __init__.py with all the imports of
        the generated class names."""
        imports = [
            Import(qname=obj.qname, source=obj.target_module)
            for obj in sorted(classes, key=lambda x: x.name)
        ]
        DependenciesResolver.resolve_conflicts(imports, set())

        output = self.env.get_template("package.jinja2").render(
            imports=imports,
            module=module,
        )
        return f"{output.strip()}\n"

    def render_module(
        self, resolver: DependenciesResolver, classes: List[Class]
    ) -> str:
        """Render the source code for the target module of the given class
        list."""

        if len({x.target_namespace for x in classes}) == 1:
            module_namespace = classes[0].target_namespace
        else:
            module_namespace = None

        resolver.process(classes)
        imports = resolver.sorted_imports()
        output = self.render_classes(resolver.sorted_classes(), module_namespace)
        module = classes[0].target_module

        return self.env.get_template("module.jinja2").render(
            output=output,
            module=module,
            imports=imports,
            namespace=module_namespace,
        )

    def render_classes(
        self, classes: List[Class], module_namespace: Optional[str]
    ) -> str:
        """Render the source code of the classes."""
        load = self.env.get_template

        def render_class(obj: Class) -> str:
            """Render class or enumeration."""

            if obj.is_enumeration:
                template = load("enum.jinja2")
            elif obj.is_service:
                template = load("service.jinja2")
            else:
                template = load("class.jinja2")

            return template.render(
                obj=obj,
                module_namespace=module_namespace,
            ).strip()

        return "\n\n\n".join(map(render_class, classes)) + "\n"

    def module_name(self, name: str) -> str:
        """Convert the given module name to safe snake case."""
        return self.filters.module_name(name)

    def package_name(self, name: str) -> str:
        """Convert the given package name to safe snake case."""
        return self.filters.package_name(name)

    @classmethod
    def ensure_packages(cls, package: Path) -> Iterator[GeneratorResult]:
        """Ensure all the __init__ files exists for the target package path,
        otherwise yield the necessary filepath, name, source output that needs
        to be crated."""
        cwd = Path.cwd()
        while cwd < package:
            init = package.joinpath("__init__.py")
            if not init.exists():
                yield GeneratorResult(
                    path=init, title="init", source="# nothing here\n"
                )
            package = package.parent

    @classmethod
    def init_filters(cls, config: GeneratorConfig) -> Filters:
        return Filters(config)
